# Vehicle.py
# 
# Created:  Apr 2024, M. Clarke
# Modified:  
 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
 
from RCAIDE                    import Framework
from RCAIDE.Framework.Core     import Data, DataOrdered
from RCAIDE.Library            import Components, Attributes 
import numpy as np 
from RCAIDE.Library.Methods.Weights.Correlation_Buildups.Common.compute_operating_empty_weight import  compute_operating_empty_weight

# ----------------------------------------------------------------------------------------------------------------------
#  Vehicle
# ----------------------------------------------------------------------------------------------------------------------  
class Vehicle(Data):
    """RCAIDE Vehicle container class with database + input / output functionality
    """    

    def __defaults__(self):
        """This sets the default values.
    
        Assumptions:
            None
        
        Source:
            None
        """    
        self.tag                                                           = 'vehicle'
        self.networks                                                      = Framework.Networks.Network.Container()
        self.fuselages                                                     = Components.Fuselages.Fuselage.Container()
        self.wings                                                         = Components.Wings.Wing.Container()
        self.nacelles                                                      = Components.Nacelles.Nacelle.Container()
        self.systems                                                       = Components.Systems.System.Container()
        self.avionics                                                      = Components.Systems.Avionics.Container()
        self.booms                                                         = Components.Booms.Boom.Container()
        self.mass_properties                                               = Vehicle_Mass_Container()
        self.payload                                                       = Components.Payloads.Payload.Container()
        self.costs                                                         = Data()     
        self.landing_gears                                                 = Components.Landing_Gear.Landing_Gear.Container()  
        self.reference_area                                                = 0.0
        self.passengers                                                    = 0.0
        self.maximum_cross_sectional_area                                  = 0.0
        
        self.flight_envelope                                               = Data()
        self.flight_envelope.design_dynamic_pressure                       = None 
        self.flight_envelope.design_mach_number                            = None  
        self.flight_envelope.design_cruise_altitude                        = None
        self.flight_envelope.design_range                                  = None 
        self.flight_envelope.ultimate_load                                 = 5.7  
        self.flight_envelope.positive_limit_load                           = 3.8
        self.flight_envelope.negative_limit_load                           = -1.5    
        self.flight_envelope.alpha_maximum                                 = 0.0
        self.flight_envelope.category                                      = None 
        self.flight_envelope.FAR_part_number                               = None
        self.flight_envelope.alt_vc                                        = 0.0
        self.flight_envelope.alt_gust                                      = 0.0
        self.flight_envelope.max_ceiling                                   = 0.0
        self.flight_envelope.maximum_dynamic_pressure                      = 0.0
        self.flight_envelope.maximum_mach_operational                      = 0.0
        self.flight_envelope.maximum_lift_coefficient                      = None
        self.flight_envelope.minimum_lift_coefficient                      = None

        self.flight_envelope.maneuver                                      = Data()
        self.flight_envelope.maneuver.load_alleviation_factor              = 0.0 
        self.flight_envelope.maneuver.equivalent_speed                     = Data()
        self.flight_envelope.maneuver.equivalent_speed.velocity_max_gust   = 0
        self.flight_envelope.maneuver.equivalent_speed.velocity_max_cruise = 0
        self.flight_envelope.maneuver.equivalent_speed.velocity_max_dive   = 0 
        self.flight_envelope.maneuver.load_factor                          = Data()
        self.flight_envelope.maneuver.load_factor.velocity_max_gust        = 0
        self.flight_envelope.maneuver.load_factor.velocity_max_cruise      = 0
        self.flight_envelope.maneuver.load_factor.velocity_max_dive        = 0

        self.flight_envelope.gust                                          = Data()
        self.flight_envelope.gust.load_alleviation_factor                  = 0.0
        self.flight_envelope.gust.equivalent_speed                         = Data()
        self.flight_envelope.gust.equivalent_speed.velocity_max_gust       = 0
        self.flight_envelope.gust.equivalent_speed.velocity_max_cruise     = 0
        self.flight_envelope.gust.equivalent_speed.velocity_max_dive       = 0
        
        self.flight_envelope.gust.load_factor                              = Data()
        self.flight_envelope.gust.load_factor.velocity_max_gust            = 0
        self.flight_envelope.gust.load_factor.velocity_max_cruise          = 0
        self.flight_envelope.gust.load_factor.velocity_max_dive            = 0 
        
        self.performance                              = DataOrdered()
         
    _energy_network_root_map = None 

    def __init__(self,*args,**kwarg):
        """ Sets up the component hierarchy for a vehicle
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            None
        """          
        # will set defaults
        super(Vehicle,self).__init__(*args,**kwarg)  

        self._component_root_map = {
            Components.Fuselages.Fuselage              : self['fuselages']        ,
            Components.Wings.Wing                      : self['wings']            ,
            Components.Systems.System                  : self['systems']          ,
            Components.Systems.Avionics                : self['avionics']         ,
            Components.Payloads.Payload                : self['payload']          , 
            Components.Nacelles.Nacelle                : self['nacelles']         , 
            Components.Booms.Boom                      : self['booms']            ,
            Components.Landing_Gear.Landing_Gear       : self['landing_gears']     ,
            Vehicle_Mass_Properties                    : self['mass_properties']  ,
        }
         
        self._energy_network_root_map= {
            Framework.Networks.Network                 : self['networks']         ,
            }    
        
        self.append_component(Vehicle_Mass_Properties())
         
        return
    

    def find_component_root(self,component):
        """ find pointer to component data root.
        
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            None
        """  

        # find component root by type, allow subclasses
        for component_type, component_root in self._component_root_map.items():
            if isinstance(component,component_type):
                break
        else:
            raise Exception("Unable to place component type %s" % component.typestring())

        return component_root


    def append_component(self,component):
        """ Adds a component to vehicle
        
            Assumptions:
                None
    
            Source:
                None
        """  

        # assert database type
        if not isinstance(component,Data):
            raise Exception('input component must be of type Data()')

        # find the place to store data
        component_root = self.find_component_root(component)
        
        # See if the component exists, if it does modify the name
        keys = component_root.keys()
        if str.lower(component.tag) in keys:
            string_of_keys = "".join(component_root.keys())
            n_comps = string_of_keys.count(component.tag)
            component.tag = component.tag + str(n_comps+1)

        # store data
        component_root.append(component)

        return

    def find_energy_network_root(self,energy_network):
        """ Find pointer to energy network data root.
        
            Assumptions:
                None
    
            Source:
                None
        """  
        # find energy network root by type, allow subclasses
        for energy_network_type, energy_network_root in self._energy_network_root_map.items():
            if isinstance(energy_network,energy_network_type):
                break
        else:
            raise Exception("Unable to place energy_network type %s" % energy_network.typestring())

        return energy_network_root

    def sum_mass(self):
        """ Regresses through the vehicle and sums the masses
        
            Assumptions:
                None
    
            Source:
                None
        """  

        total = 0.0
        
        for key in self.keys():
            item = self[key]
            if isinstance(item,Components.Component.Container):
                total += item.sum_mass()

        return total
    
    
    def center_of_gravity(self):
        """ will recursively search the data tree and sum
            any Comp.Mass_Properties.mass, and return the total sum
        
            Assumptions:
                None
    
            Source:
                None
        """   
        total = np.array([[0.0,0.0,0.0]])

        for key in self.keys():
            item = self[key]
            if isinstance(item,Components.Component.Container):
                total += item.total_moment()
                
        mass = self.sum_mass()
        if mass ==0:
            mass = 1.
                
        CG = total/mass
        
        self.mass_properties.center_of_gravity = CG
                
        return CG
    
    def moment_of_inertia(self):
        """  
        """
        M = np.zeros((3, 3))
        center_of_gravity =  self.mass_properties.center_of_gravity 
        for key in self.keys():
            item = self[key] 
            M += item.moment_of_inertia(center_of_gravity) 
                
        return M
     
    def operating_empty_weight(self):
        """ Compute operating empty weight  
        
            Assumptions:
                None
    
            Source:
                None
        """  
        outputs = compute_operating_empty_weight(self)  
                
        return outputs     
    
    def append_energy_network(self,energy_network):
        """ Adds an energy network to vehicle 
        
            Assumptions:
                None
    
            Source:
                None
        """  

        # assert database type
        if not isinstance(energy_network,Data):
            raise Exception('input energy network must be of type Data()')

        # find the place to store data
        energy_network_root = self.find_energy_network_root(energy_network)
        
        # See if the energy network exists, if it does modify the name
        keys = energy_network_root.keys()
        if str.lower(energy_network.tag) in keys:
            string_of_keys = "".join(energy_network_root.keys())
            n_comps = string_of_keys.count(energy_network.tag)
            energy_network.tag = energy_network.tag + str(n_comps+1)

        # store data
        energy_network_root.append(energy_network)

        return    

    def find_energy_network_root(self,energy_network):
        """ Find pointer to energy network data root.
        
            Assumptions:
                None
    
            Source:
                None
        """  
        # find energy network root by type, allow subclasses
        for energy_network_type, energy_network_root in self._energy_network_root_map.items():
            if isinstance(energy_network,energy_network_type):
                break
        else:
            raise Exception("Unable to place energy_network type %s" % energy_network.typestring())

        return energy_network_root
 
class Vehicle_Mass_Properties(Components.Mass_Properties): 
    """ The vehicle's mass properties.
        
            Assumptions:
                None
    
            Source:
                None
    """

    def __defaults__(self):
        """This sets the default values.
        
            Assumptions:
                None
    
            Source:
                None
            """         

        self.tag                         = 'mass_properties'
        self.operating_empty             = 0.0
        self.max_takeoff                 = 0.0
        self.takeoff                     = 0.0
        self.max_landing                 = 0.0
        self.landing                     = 0.0
        self.max_cargo                   = 0.0
        self.cargo                       = 0.0
        self.max_payload                 = 0.0
        self.payload                     = 0.0
        self.passenger                   = 0.0
        self.crew                        = 0.0
        self.max_fuel                    = 0.0
        self.fuel                        = 0.0
        self.max_zero_fuel               = 0.0
        self.center_of_gravity           = [[0.0,0.0,0.0]]
        self.zero_fuel_center_of_gravity = np.array([[0.0,0.0,0.0]])    
        
class Vehicle_Mass_Container(Components.Component.Container,Vehicle_Mass_Properties):
        
    def append(self,value,key=None):
        """ Appends the vehicle mass, but only let's one ever exist. Keeps the newest one
        
            Assumptions:
                None
    
            Source:
                None
        """      
        self.clear()
        for key in value.keys():
            self[key] = value[key]

    def get_children(self):
        """ Returns the components that can go inside
        
            Assumptions:
                None
    
            Source:
                None
        """       
        
        return [Vehicle_Mass_Properties]



# RCAIDE/Library/Methods/Aerodynamics/Athena_Vortex_Lattice/AVL_Objects/Body.py
#
# Created: Oct 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports
from RCAIDE.Framework.Core import  Data  

# ----------------------------------------------------------------------------------------------------------------------
#  Body
# ----------------------------------------------------------------------------------------------------------------------  
class Body(Data):
    """ A data class defining the parameters of a fuselage or other body modeled
    by side and planform projections arranged in a plus (+) shape (when viewed from
    the front).

    Assumptions:
        None

    Source:
        None

    Inputs:
        None

    Outputs:
        None

    Properties Used:
        N/A
    """    


    def __defaults__(self):
        """ Defines the data structure and defaults for aircraft body geometry 

        Assumptions:
            None

        Source:
            None

        Inputs:
            None

        Outputs:
            None

        Properties Used:
            N/A
        """ 		
        self.tag                      = 'body'
        self.symmetric                = True
        self.origin                   = [0.,0.,0.]

        self.lengths                  = Data()
        self.lengths.total            = 0.0
        self.lengths.nose             = 0.0
        self.lengths.tail             = 0.0

        self.widths                   = Data()
        self.widths.maximum           = 0.0
        self.heights                  = Data()
        self.heights.maximum          = 0.0

        self.sections                 = Data()
        self.sections.vertical        = Data()
        self.sections.horizontal      = Data()
        self.configuration            = Data()

        self.configuration.nspanwise  = 10
        self.configuration.nchordwise = 5
        self.configuration.sspace     = 1.0
        self.configuration.cspace     = 1.0


    def append_section(self,section,orientation='Horizontal'):
        """ Adds a section to the body vertical or horizontal segment 

        Assumptions:
            None

        Source:
            None

        Inputs:
            None

        Outputs:
            None

        Properties Used:
            N/A
        """ 
        # assert database type
        if not isinstance(section,Data):
            raise Exception('input component must be of type Data()')

        # store data
        if orientation.lower() == 'horizontal':
            self.sections.horizontal.append(section)
        elif orientation.lower() == 'vertical':
            self.sections.vertical.append(section)
        else:
            raise KeyError('No key, "{}". Use "Horizontal" or "Vertical".'.format(orientation))
        return



# RCAIDE/Library/Methods/Weights/Buildups/Common/compute_boom_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
 
# package imports 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
# Compute wiring weight
# ----------------------------------------------------------------------------------------------------------------------
def compute_wiring_weight(wing, config, cablePower):
    """ Calculates mass of wiring required for a wing, including DC power
        cables and communication cables, assuming power cables run an average of
        half the fuselage length and height in addition to reaching the motor
        location on the wingspan, and that communication and sesor  wires run an
        additional length based on the fuselage and wing dimensions. 
        
        Sources:
        Project Vahana Conceptual Trade Study

        Inputs:

            config                      RCAIDE Config Data Structure 
            max_power_draw              Maximum DC Power Draw           [W]

        Outputs:

            weight:                     Wiring Mass                     [kg]

    """
    weight      = 0.0 
    cableLength = 0.0
    for network in config.networks:
        for propulsor in network.propulsors:
            motor = propulsor.motor
            if propulsor.wing_mounted == True:  
                MSL             = np.array(motor.origin) #- np.array(bus.origin)  WHAT DO WE DO HERE SINCE WE DONT HAVE A BUS DEFINED YET AT THIS STAGE
                cableLength     += np.sum(abs(MSL))  
                    
    cableDensity    = 5.7e-6
    massCables      = cableDensity * cablePower * cableLength
     
    # Determine mass of sensor/communication wires
    
    fLength = 0
    for fus in config.fuselages:
        fLength  += fus.lengths.total 
    
    wiresPerBundle  = 6
    wireDensity     = 460e-5
    wireLength      = cableLength + (10 * fLength) +  4*wing.spans.projected
    massWires       = wireDensity * wiresPerBundle * wireLength
     
    # Sum Total 
    weight += massCables + massWires
    
    return weight


# RCAIDE/Library/Components/Propulsors/Converters/Expansion_Nozzle.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Feb 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
 # RCAIDE imports   
from RCAIDE.Library.Components                      import Component   
from RCAIDE.Library.Methods.Propulsors.Converters.Expansion_Nozzle.append_expansion_nozzle_conditions import append_expansion_nozzle_conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  Expansion Nozzle
# ----------------------------------------------------------------------------------------------------------------------
class Expansion_Nozzle(Component):
    """
    An expansion nozzle component model for propulsion systems that converts pressure energy into kinetic energy.

    Attributes
    ----------
    tag : str
        Identifier for the nozzle. Default is 'Nozzle'.
        
    polytropic_efficiency : float
        Efficiency of the expansion process accounting for losses. Default is 1.0.
        
    pressure_ratio : float
        Ratio of outlet to inlet pressure. Default is 1.0.

    Notes
    -----
    The Expansion_Nozzle class models a nozzle that converts pressure energy into 
    kinetic energy through an expansion process. The model:
    * Handles subsonic and choked flow conditions
    * Accounts for losses through polytropic efficiency
    * Maintains constant pressure ratio at design conditions
    * Assumes quasi-one-dimensional flow

    **Major Assumptions**
    * Pressure ratio and efficiency do not change with varying conditions
    * Flow can be subsonic or choked (but not supersonic)

    **Definitions**

    'Polytropic Efficiency'
        Measure of expansion process efficiency accounting for real gas effects

    'Pressure Ratio'
        Ratio of exit to inlet pressure
    
    'Choked Flow'
        Condition where flow reaches sonic velocity at the throat

    References
    ----------
    [1] Mattingly, J. D., & Boyer, K. M. (2016). Elements of propulsion: Gas 
        turbines and rockets, second edition Jack D. Mattingly, Keith M. Boyer. 
        American Institute of Aeronautics and Astronautics.

    See Also
    --------
    RCAIDE.Library.Components.Component
    RCAIDE.Library.Components.Propulsors.Converters.Compression_Nozzle
    RCAIDE.Library.Methods.Propulsors.Converters.Expansion_Nozzle
    """
    
    def __defaults__(self):
        """This sets the default values for the component to function.

        Assumptions:
            1. Pressure ratio and efficiency do not change with varying conditions.
            2. Subsonic or choked output.

        Source:
            https://web.stanford.edu/~cantwell/AA283_Course_Material/AA283_Course_Notes/
        """          
        #set the defaults
        self.tag = 'Nozzle'
        self.polytropic_efficiency           = 1.0
        self.pressure_ratio                  = 1.0

    def append_operating_conditions(self,segment,propulsor): 
        propulsor_conditions =  segment.state.conditions.energy[propulsor.tag]
        append_expansion_nozzle_conditions(self,segment,propulsor_conditions)
        return                        


# RCAIDE/Library/Methods/Propulsors/Electric_Rotor_Propulsor/append_electric_ducted_fan_conditions.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
 # RCAIDE imports  
from RCAIDE.Framework.Mission.Common                             import Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append electric ducted fan network conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_electric_ducted_fan_conditions(propulsor,segment):
    '''
    
    
    
    '''
    ones_row    = segment.state.ones_row
                
    segment.state.conditions.energy[propulsor.tag]                               = Conditions()  
    segment.state.conditions.energy[propulsor.tag].throttle                      = 0. * ones_row(1)      
    segment.state.conditions.energy[propulsor.tag].commanded_thrust_vector_angle = 0. * ones_row(1)  
    segment.state.conditions.energy[propulsor.tag].thrust                        = 0. * ones_row(3) 
    segment.state.conditions.energy[propulsor.tag].power                         = 0. * ones_row(1) 
    segment.state.conditions.energy[propulsor.tag].moment                        = 0. * ones_row(3)         
    return


# RCAIDE/Methods/Noise/Correlation_Buildup/Airframe/clean_wing_noise.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------   
import numpy as np
from RCAIDE.Framework.Core import Units

# ----------------------------------------------------------------------------------------------------------------------  
# main and nose landing gear noise
# ----------------------------------------------------------------------------------------------------------------------  
def landing_gear_noise(D,H,wheels,M,velocity,phi,theta,distance,frequency):
    """This calculates the Landing gear 1/3 octave band sound pressure level and overall sound pressure level
    for a tyre diameter D, a strut length H and WHEELS number of  wheels per unit.

    Assumptions:
        Correlation based.

    Source:
        None


    Inputs:
        D         - Landing gear tyre diameter                                     [ft]
        H         - Lading gear strut length                                       [ft]
        wheels    - Number of wheels per unit                                      [-]
        M         - Mach number                                                    [-]
        velocity  - Aircraft speed                                                 [kts]
        phi       - Azimuthal angle                                                [rad]
        theta     - Polar angle                                                    [rad]
        distance  - Distance from airplane to observer, evaluated at retarded time [ft]
        frequemcy - Frequency array                                                [Hz] 

    Outputs: One Third Octave Band SPL [dB]
        SPL           - Sound Pressure Level of the landing gear         [dB]
        OASPL         - Overall Sound Pressure Level of the landing gear [dB]


    Source:
        SAE ARP 876D

    Inputs:
        Ce = wing chord length at the engine location - as figure 7.3 of the SAE ARP 876D                    [m]                          
        Xe = fan exit location downstream of the leading edge (Xe<Ce) - as figure 7.3 of the SAE ARP 876D    [m] 
        Ye = separation distance from the wing chord line to nozzle lip - as figure 7.3 of the SAE ARP 876D  [m] 
        theta_s                                                                                              [rad]
        Diameter_mixed                                                                                       [m] 

    Outputs:
        INST_s          [-]

    Properties Used: 
        N/A 
        
    """  
    
    velocity_fts = velocity/Units.ft
    velocity_kts = velocity/Units.knots

    if (wheels==1 or wheels==2):
        G1 = 13+np.log10(4.5*((frequency*D/(velocity_fts*(1-M*np.cos(theta))))**2)* \
            (12.5+((frequency*D/(velocity_fts*(1-M*np.cos(theta))))**2))**-2.25)
        G2 = (13+np.log10(2.0*(frequency*D/(velocity_fts*(1-M*np.math.cos(theta)))**2.0))* \
            (30+(frequency*D/(velocity_fts*(1-M*np.cos(theta))))**8)**-1*(0.34*H/D))* \
            (np.math.sin(phi))**2
    elif wheels==4:
        G1 = 12+np.log10(frequency*D/(velocity_fts*(1-M*np.cos(theta))))**2 \
        *(0.4+(frequency*D/(velocity_fts*(1-M*np.cos(theta))))**2)**(-1.6)
        G2 = (12+np.log10(7.0*(frequency*D/(velocity_fts*(1-M*np.cos(theta))))**3.0 * \
            (1.06+(frequency*D/(velocity_fts*(1-M*np.cos(theta))))**2)**(-3.0)*(1)))*(np.sin(phi))**2


    G3    = 12.79+np.log10(0.34*H/D)*(np.sin(phi))**2
    SPL   = 60.*np.log10(velocity_kts/194.0)+20.*np.log10(D/distance)+10.*np.log10(10.0**G1+10.0**G2)
    OASPL = 60.*np.log10(velocity_kts/194.0)+20.*np.log10(D/distance)+10.*np.log10(10.0**12.52+10.0**G3)

    return SPL



# RCAIDE/Methods/Energy/Propulsors/Rotor_Design/set_optimized_parameters.py
# 
# 
# Created:  Jul 2023, M. Clarke 
    
# ----------------------------------------------------------------------------------------------------------------------  
#  Set Optimized rotor platform paramters
# ----------------------------------------------------------------------------------------------------------------------   
def set_optimized_parameters(rotor,optimization_problem):
    """ Append parameters of optimized prop-rotor to input prop-rotor
          
          Inputs:  
             rotor                - rotor data structure                        [None]
             optimization_problem      - data struction of optimized parameters [None]
             
          Outputs: 
             rotor                - rotor data structure                        [None]
              
          Assumptions: 
             1) Default noise measurements are taken 135 degrees from prop-rotor plane 
        
          Source:
             None
    """    
    results                         = optimization_problem.results 
    optimal_hover_rotor             = optimization_problem.vehicle_configurations.hover.networks.electric.propulsors.electric_rotor.rotor     
    optimal_oei_rotor               = optimization_problem.vehicle_configurations.oei.networks.electric.propulsors.electric_rotor.rotor    
    rotor.chord_distribution        = optimal_hover_rotor.chord_distribution
    rotor.twist_distribution        = optimal_hover_rotor.twist_distribution   
    
    if rotor.hover.design_power == None: 
        rotor.hover.design_power = results.hover.power 
    
    if rotor.hover.design_thrust == None: 
        rotor.hover.design_thrust = results.hover.thrust   
        
    rotor.hover.design_torque              = results.hover.torque  
    rotor.hover.design_angular_velocity    = results.hover.omega  
    rotor.hover.design_Cl                  = results.hover.mean_CL 
    rotor.hover.design_thrust_coefficient  = results.hover.thurst_c
    rotor.hover.design_power_coefficient   = results.hover.power_c 
    rotor.hover.design_SPL_dBA             = results.hover.mean_SPL 
    rotor.hover.design_pitch_command       = optimal_hover_rotor.hover.design_pitch_command
    rotor.oei.design_thrust                = results.oei.thrust    
    rotor.oei.design_power                 = results.oei.power 
    rotor.oei.design_torque                = results.oei.torque  
    rotor.oei.design_angular_velocity      = results.oei.omega 
    rotor.oei.design_pitch_command         = optimal_oei_rotor.oei.design_pitch_command
    
    if optimization_problem.prop_rotor_flag:  
        optimal_cruise_rotor  = optimization_problem.vehicle_configurations.cruise.networks.electric.propulsors.electric_rotor.rotor      
        if rotor.cruise.design_power == None: 
            rotor.cruise.design_power = results.cruise.power 
        
        if rotor.cruise.design_thrust == None: 
            rotor.cruise.design_thrust = results.cruise.thrust            
     
        rotor.cruise.design_torque              = results.cruise.torque  
        rotor.cruise.design_angular_velocity    = results.cruise.omega  
        rotor.cruise.design_Cl                  = results.cruise.mean_CL 
        rotor.cruise.design_thrust_coefficient  = results.cruise.thurst_c
        rotor.cruise.design_power_coefficient   = results.cruise.power_c   
        rotor.cruise.design_SPL_dBA             = results.cruise.mean_SPL
        rotor.cruise.design_pitch_command       = optimal_cruise_rotor.cruise.design_pitch_command
             
    rotor.max_thickness_distribution        = optimal_hover_rotor.max_thickness_distribution  
    rotor.radius_distribution               = optimal_hover_rotor.radius_distribution         
    rotor.number_of_blades                  = optimal_hover_rotor.number_of_blades               
    rotor.mid_chord_alignment               = optimal_hover_rotor.mid_chord_alignment         
    rotor.thickness_to_chord                = optimal_hover_rotor.max_thickness_distribution / optimal_hover_rotor.chord_distribution
    rotor.blade_solidity                    = optimal_hover_rotor.blade_solidity   
    
    return rotor 


# RCAIDE/Methods/Noise/Correlation_Buildup/Engine/primary_noise_component.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
 
# Python package imports   
import numpy as np   

# ----------------------------------------------------------------------------------------------------------------------     
#  Primary Noise Component
# ----------------------------------------------------------------------------------------------------------------------          
def primary_noise_component(Velocity_primary,Temperature_primary,R_gas,theta_p,DVPS,sound_ambient,Velocity_secondary,Velocity_aircraft,Area_primary,Area_secondary,DSPL_p,EX_p,Str_p):
    """This function calculates the noise contribution of the primary jet component
    
        Assumptions:
        Empirical based procedure.
    
    Source: 
        None
        
    Inputs:
        noise_data     - RCAIDE type vehicle

    Outputs:
        OASPL          - Overall Sound Pressure Level            [dB]
        PNL            - Perceived Noise Level                   [dB]
        PNL_dBA        - Perceived Noise Level A-weighted level  [dBA]
        EPNdB_takeoff  - Takeoff Effective Perceived Noise Level [EPNdB]
        EPNdB_landing  - Landing Effective Perceived Noise Level [EPNdB]  
    
    Properties Used:
        N/A  
       
    """      

    # Flow parameters of the primary jet
    sound_primary    = np.sqrt(1.4*R_gas*Temperature_primary) 
    Mach_primary_jet = Velocity_primary/sound_primary  
    
    # Calculation of the velocity exponent 
    velocity_exponent = 1.5*np.exp(-10*(theta_p - 2.2)**2)
    if theta_p <= 2.2: 
        velocity_exponent = 1.56

    # Calculation of the Source Strengh Function (FV)
    FV = Mach_primary_jet*(DVPS/sound_ambient)**0.6*((Velocity_primary+Velocity_secondary)/sound_ambient)**0.4* \
    (np.abs(Velocity_primary-Velocity_aircraft)/Velocity_primary)**velocity_exponent

    # Determination of the noise model coefficients
    Z1 = -18*((1.8*theta_p/np.pi)-0.6)**2
    Z2 = -18-18*((1.8*theta_p/np.pi)-0.6)**2
    Z3 = 0.0
    Z4 = -0.1 - 0.75*((Velocity_primary-Velocity_secondary-Velocity_aircraft)/sound_ambient) * \
        ((1.8*theta_p/np.pi)-0.6)**3. + 0.8*(0.6-np.log10(1+Area_secondary/Area_primary))
    Z5 = 50 + 20*np.exp(-(theta_p-2.6)**2.)
    Z6 = 94 + 46*np.exp(-(theta_p-2.5)**2.) - 26.*(0.6-np.log10(1+Area_secondary/Area_primary))/ \
        np.exp(5*(theta_p-2.3)**2) + DSPL_p + EX_p

    # Determination of Sound Pressure Level for the primary jet component
    SPL_p = (Z1*np.log10(FV)+Z2) * (np.log10(Str_p)-Z3*np.log10(FV)-Z4)**2 + Z5*np.log10(FV) + Z6

    return SPL_p




# RCAIDE/Library/Plots/Performance/Stability/plot_longitudinal_stability.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np  

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------   
def plot_longitudinal_stability(results,
                             save_figure = False,
                             show_legend=True,
                             save_filename = "Longitudinal_Stability",
                             file_type = ".png",
                             width = 11, height = 7):
    """This plots the static stability characteristics of an aircraft 
    """  

    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters)
     
    # get line colors for plots 
    line_colors   = cm.inferno(np.linspace(0,0.9,len(results.segments)))     
     
    fig   = plt.figure(save_filename)
    fig.set_size_inches(width,height)
    

    axis_1 = plt.subplot(3,2,1)    
    axis_2 = plt.subplot(3,2,2) 
    axis_3 = plt.subplot(3,2,3)  
    axis_4 = plt.subplot(3,2,4)  
    axis_5 = plt.subplot(3,2,5)
    axis_6 = plt.subplot(3,2,6) 
    
    for i in range(len(results.segments)): 
        time       = results.segments[i].conditions.frames.inertial.time[:,0] / Units.min
        c_m        = results.segments[i].conditions.static_stability.coefficients.M[:,0]   
        SM         = results.segments[i].conditions.static_stability.static_margin[:,0]  
        delta_e    = results.segments[i].conditions.control_surfaces.elevator.deflection[:,0] / Units.deg
        CM_delta_e = results.segments[i].conditions.static_stability.derivatives.CM_delta_e[:,0]
        Cm_alpha   = results.segments[i].conditions.static_stability.derivatives.CM_alpha[:,0]
        CL_alpha   = results.segments[i].conditions.static_stability.derivatives.Clift_alpha[:,0] 
          
        segment_tag  =  results.segments[i].tag
        segment_name = segment_tag.replace('_', ' ')  
        
        axis_1.plot(time, c_m, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width, label = segment_name) 
        axis_1.set_ylabel(r'$C_M$')
        axis_1.set_ylim([-1, 1])  
        set_axes(axis_1) 

        axis_2.plot(time, Cm_alpha, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width) 
        axis_2.set_ylabel(r'$C_M\alpha$')
        axis_2.set_ylim([-1, 1]) 
        set_axes(axis_2) 
        
        axis_3.plot(time,SM , color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width)
        axis_3.set_xlabel('Time (mins)')
        axis_3.set_ylabel(r'Static Margin (%)')
        set_axes(axis_3)  

        axis_4.plot(time,delta_e , color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width)
        axis_4.set_xlabel('Time (mins)')
        axis_4.set_ylabel(r'Elevator Defl.n') 
        axis_4.set_ylim([-15, 15]) 
        set_axes(axis_4) 
        
        axis_5.plot(time,CM_delta_e , color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width)
        axis_5.set_xlabel('Time (mins)')
        axis_5.set_ylabel(r'$C_M\delta_e$')
        axis_5.set_ylim([-1, 1]) 
        set_axes(axis_5)
        
        axis_6.plot(time,CL_alpha, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width)
        axis_6.set_xlabel('Time (mins)')
        axis_6.set_ylabel(r'$C_L\alpha$')
        axis_6.set_ylim([-1, 10]) 
        set_axes(axis_6)    
        
    if show_legend:
        leg =  fig.legend(bbox_to_anchor=(0.5, 0.95), loc='upper center', ncol = 4) 
        leg.set_title('Flight Segment', prop={'size': ps.legend_font_size, 'weight': 'heavy'})     
          
    # Adjusting the sub-plots for legend
    fig.tight_layout()
    fig.subplots_adjust(top=0.8)
    
    # set title of plot 
    title_text    = 'Stability Coefficents'      
    fig.suptitle(title_text)
 
    if save_figure:
        plt.savefig(save_filename + file_type)   
    return fig 
 


# RCAIDE/Library/Plots/Topography/plot_elevation_contours.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------   
from RCAIDE.Framework.Core                             import Units
from RCAIDE.Framework.Analyses.Geodesics.Geodesics import Calculate_Distance
from RCAIDE.Library.Plots.Common import plot_style

# python imports 
import matplotlib.pyplot as plt
from scipy.interpolate import griddata
import matplotlib.colors 
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------    
def plot_elevation_contours(topography_file,
                            number_of_latitudinal_points  = 100,
                            number_of_longitudinal_points = 100, 
                            use_lat_long_coordinates      = True, 
                            save_figure = False,  
                            show_legend = True,
                            save_filename = "Elevation_Contours",
                            file_type = ".png",
                            width = 11, height = 7): 
    

    """This plots the elevation contours

    Assumptions:
    None

    Source:
    topography_file

    Inputs:
    vehicle

    Outputs:
    Plots

    Properties Used:
    N/A
    """

    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters)
     
     
    colors_undersea = plt.cm.terrain(np.linspace(0, 0.17, 56))
    colors_land     = plt.cm.terrain(np.linspace(0.25, 1, 200)) 
    
    # combine them and build a new colormap
    colors          = np.vstack((colors_undersea, colors_land))
    cut_terrain_map = matplotlib.colors.LinearSegmentedColormap.from_list('cut_terrain', colors) 
    
    data = np.loadtxt(topography_file)
    Long = data[:,0]
    Lat  = data[:,1]
    Elev = data[:,2]    

    x_min_coord = np.min(Lat)
    x_max_coord = np.max(Lat)
    y_min_coord = np.min(Long)
    y_max_coord = np.max(Long)
    if np.min(Long)>180: 
        y_min_coord = np.min(Long)-360
    if np.max(Long)>180:
        y_max_coord = np.max(Long)-360  
    
    top_left_map_coords      = np.array([x_max_coord,y_min_coord])
    bottom_left_map_coords   = np.array([x_min_coord,y_min_coord]) 
    top_right_map_coords     = np.array([x_max_coord,y_max_coord])
    bottom_right_map_coords  = np.array([x_min_coord,y_max_coord]) 
    
    x_dist_max = Calculate_Distance(top_left_map_coords,bottom_left_map_coords) * Units.kilometers
    y_dist_max = Calculate_Distance(bottom_right_map_coords,bottom_left_map_coords) * Units.kilometers
    
    [long_dist,lat_dist]  = np.meshgrid(np.linspace(0,y_dist_max,number_of_longitudinal_points),np.linspace(0,x_dist_max,number_of_latitudinal_points))
    [long_deg,lat_deg]    = np.meshgrid(np.linspace(np.min(Long),np.max(Long),number_of_longitudinal_points),np.linspace(np.min(Lat),np.max(Lat),number_of_latitudinal_points)) 
    elevation             = griddata((Lat,Long), Elev, (lat_deg, long_deg), method='linear')     
    elevation             = elevation/Units.feet
    norm                  = FixPointNormalize(sealevel=0,vmax=np.max(elevation),vmin=np.min(elevation)) 
    
    fig = plt.figure(save_filename)
    fig.set_size_inches(width,height)
    axis = fig.add_subplot(1,1,1) 
    
    if use_lat_long_coordinates:
        CS   = axis.contourf(long_deg,lat_deg,elevation,cmap =cut_terrain_map,norm=norm,levels = 20)   
        cbar = fig.colorbar(CS, ax=axis)     
        cbar.ax.set_ylabel('Elevation above sea level [ft]', rotation =  90)  
        axis.set_xlabel('Longitude [°]')
        axis.set_ylabel('Latitude [°]') 
    else: 
        CS   = axis.contourf(long_dist/Units.nmi,lat_dist/Units.nmi,elevation,cmap =cut_terrain_map,norm=norm,levels = 20)  
        cbar = fig.colorbar(CS, ax=axis)        
        cbar.ax.set_ylabel('Elevation above sea level [ft]', rotation =  90) 
        axis.set_xlabel('Longitudinal Distance [nmi]')
        axis.set_ylabel('Latitudinal Distance [nmi]') 
     
    return  fig   

class FixPointNormalize(matplotlib.colors.Normalize):
    """ 
    Inspired by https://stackoverflow.com/questions/20144529/shifted-colorbar-matplotlib
    Subclassing Normalize to obtain a colormap with a fixpoint 
    somewhere in the middle of the colormap.
    This may be useful for a `terrain` map, to set the "sea level" 
    to a color in the blue/turquise range. 
    """
    def __init__(self, vmin=None, vmax=None, sealevel=0, col_val = 0.21875, clip=False):
        # sealevel is the fix point of the colormap (in data units)
        self.sealevel = sealevel
        # col_val is the color value in the range [0,1] that should represent the sealevel.
        self.col_val = col_val
        matplotlib.colors.Normalize.__init__(self, vmin, vmax, clip)

    def __call__(self, value, clip=None):
        x, y = [self.vmin, self.sealevel, self.vmax], [0, self.col_val, 1]
        return np.ma.masked_array(np.interp(value, x, y)) 
    
  


# RCAIDE/Library/Methods/Weights/Correlations/FLOPS/__init__.py
# 

"""RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .compute_fuselage_weight            import compute_fuselage_weight          
from .compute_horizontal_tail_weight     import compute_horizontal_tail_weight  
from .compute_operating_items_weight     import compute_operating_items_weight   
from .compute_landing_gear_weight        import compute_landing_gear_weight      
from .compute_payload_weight             import compute_payload_weight           
from .compute_propulsion_system_weight   import compute_propulsion_system_weight 
from .compute_systems_weight             import compute_systems_weight           
from .compute_vertical_tail_weight       import compute_vertical_tail_weight     
from .compute_wing_weight                import compute_wing_weight              


# RCAIDE/Library/Missions/Segments/Cruise/Curved_Constant_Radius_Constant_Speed_Constant_Altitude/initialize_conditions.py
# 
# 
# Created:  September 2024, A. Molloy + M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# Package imports 
import numpy as np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------
def initialize_conditions(segment):
    """Sets the specified conditions which are given for the segment type.

    Assumptions:
    Curved segment with constant radius, constant speed and constant altitude
    Assumes that it is a coordinated turn (true course and true heading are aligned)

    Source:
    N/A

    Inputs:
    segment.altitude                [meters] # deleted segment distance as it is now defined by the turn_angle and radius.
    segment.speed                   [meters/second]
    self.start_true_course          [degrees] true course of the vehicle before the turn
    self.turn_angle                 [degrees] angle measure of the curve. + is right hand turn, - is left hand turn. 
    self.radius                     [meters] radius of the turn
    
    Outputs: ***pretty sure that no additional outputs are need. Verify this***
    conditions.frames.inertial.velocity_vector  [meters/second]
    conditions.frames.inertial.position_vector  [meters]
    conditions.freestream.altitude              [meters]
    conditions.frames.inertial.time             [seconds]

    Properties Used:
    N/A
    """        
    
    # unpack 
    alt               = segment.altitude
    air_speed         = segment.air_speed       
    beta              = segment.sideslip_angle
    radius            = segment.turn_radius
    start_true_course = segment.true_course
    arc_sector        = segment.turn_angle
    conditions        = segment.state.conditions 

    # check for initial velocity
    if air_speed is None: 
        if not segment.state.initials: raise AttributeError('airspeed not set')
        air_speed = np.linalg.norm(segment.state.initials.conditions.frames.inertial.velocity_vector[-1])
        
    # check for initial altitude
    if alt is None:
        if not segment.state.initials: raise AttributeError('altitude not set')
        alt = -1.0 * segment.state.initials.conditions.frames.inertial.position_vector[-1,2]
    
    # check for turn radius
    if radius is None:
        if not segment.state.initials: raise AttributeError('radius not set')
        radius = 0.1 # minimum radius so as to approximate a near instantaneous curve
    
    # check for turn angle
    if arc_sector is None:
        if not segment.state.initials: raise AttributeError('turn angle not set')
        arc_sector = 0.0 # aircraft does not turn    

    # dimensionalize time
    v_body_x    = np.cos(beta)*air_speed # x-velocity in the body frame. 
    v_body_y    = np.sin(beta)*air_speed # y-velocity in the body frame
    t_initial   = conditions.frames.inertial.time[0,0]
    omega       = v_body_x / radius
    t_final     = abs(arc_sector) / omega + t_initial  # Time to complete the turn
    t_nondim    = segment.state.numerics.dimensionless.control_points
    time        = t_nondim * (t_final-t_initial) + t_initial
    
    true_course_control_points = start_true_course + t_nondim * arc_sector
    
    v_inertial_x = air_speed * np.cos(true_course_control_points)
    v_inertial_y = air_speed * np.sin(true_course_control_points)
    
    # pack
    segment.state.conditions.freestream.altitude[:,0]             = alt
    segment.state.conditions.frames.inertial.position_vector[:,2] = -alt # z points down
    segment.state.conditions.frames.inertial.velocity_vector[:,0] = v_inertial_x[:,0]
    segment.state.conditions.frames.inertial.velocity_vector[:,1] = v_inertial_y[:,0]
    segment.state.conditions.frames.body.velocity_vector[:,0]     = v_body_x
    segment.state.conditions.frames.body.velocity_vector[:,1]     = v_body_y
    segment.state.conditions.frames.inertial.time[:,0]            = time[:,0]
    segment.state.conditions.frames.planet.true_heading[:,0]      = true_course_control_points[:,0]
    segment.state.conditions.frames.planet.true_course[:,0]       = true_course_control_points[:,0]


# RCAIDE/Library/Missions/Segments/Climb/Constant_EAS_Constant_Rate.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE 
from RCAIDE.Library.Mission.Common.Update.atmosphere import atmosphere

# Package imports  
import numpy as np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------
def initialize_conditions(segment):
    """Sets the specified conditions which are given for the segment type.
    
    Assumptions:
    Constant true airspeed with a constant rate of climb

    Source:
    N/A

    Inputs:
    segment.climb_rate                                  [meters/second]
    segment.equivalent_air_speed                        [meters/second]
    segment.altitude_start                              [meters]
    segment.altitude_end                                [meters]
    segment.state.numerics.dimensionless.control_points [Unitless]
    conditions.freestream.density                       [kilograms/meter^3]

    Outputs:
    conditions.frames.inertial.velocity_vector  [meters/second]
    conditions.frames.inertial.position_vector  [meters]
    conditions.freestream.altitude              [meters]

    Properties Used:
    N/A
    """         
    
    # unpack
    climb_rate = segment.climb_rate
    eas        = segment.equivalent_air_speed    
    beta       = segment.sideslip_angle
    alt0       = segment.altitude_start 
    altf       = segment.altitude_end
    t_nondim   = segment.state.numerics.dimensionless.control_points
    conditions = segment.state.conditions  

    # check for initial altitude
    if alt0 is None:
        if not segment.state.initials: raise AttributeError('initial altitude not set')
        alt0 = -1.0 * segment.state.initials.conditions.frames.inertial.position_vector[-1,2]

    # discretize on altitude
    alt = t_nondim * (altf-alt0) + alt0
    
    # pack conditions
    conditions.freestream.altitude[:,0]             =  alt[:,0] # positive altitude in this context


    # check for initial velocity vector
    if eas is None:
        if not segment.state.initials: raise AttributeError('initial equivalent airspeed not set')
        air_speed  = np.linalg.norm(segment.state.initials.conditions.frames.inertial.velocity_vector[-1,:])  
    else: 
        # determine airspeed from equivalent airspeed
        atmosphere(segment) # get density for airspeed
        density   = conditions.freestream.density[:,0]   
        MSL_data  = segment.analyses.atmosphere.compute_values(0.0,0.0)
        air_speed = eas/np.sqrt(density/MSL_data.density[0])    
    
    # process velocity vector
    v_mag  = air_speed
    v_z    = -climb_rate # z points down
    v_xy   = np.sqrt( v_mag**2 - v_z**2 )
    v_x    = np.cos(beta)*v_xy
    v_y    = np.sin(beta)*v_xy
    
    # pack conditions    
    conditions.frames.inertial.velocity_vector[:,0] = v_x
    conditions.frames.inertial.velocity_vector[:,1] = v_y
    conditions.frames.inertial.velocity_vector[:,2] = v_z
    conditions.frames.inertial.position_vector[:,2] = -alt[:,0] # z points down


# RCAIDE/Library/Compoments/Wings/Wing.py
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports
import RCAIDE
from RCAIDE.Framework.Core      import Data,Container 
from RCAIDE.Library.Components  import Mass_Properties, Component   
from RCAIDE.Library.Methods.Weights.Moment_of_Inertia.compute_wing_moment_of_inertia import  compute_wing_moment_of_inertia

import numpy as np

# ---------------------------------------------------------------------------------------------------------------------- 
#  Wing
# ----------------------------------------------------------------------------------------------------------------------  
class Wing(Component):
    """
    Base class for aircraft lifting surfaces providing core functionality for geometric 
    definition and analysis.

    Attributes
    ----------
    tag : str
        Unique identifier for the wing, defaults to 'wing'
        
    mass_properties : Mass_Properties
        Mass and inertia properties, initialized empty
        
    origin : ndarray
        3D coordinates [x, y, z] defining wing's reference point, defaults to [0.0, 0.0, 0.0]
        
    symmetric : bool
        Flag indicating if wing is symmetric about x-z plane, defaults to True
        
    vertical : bool
        Flag indicating if wing is vertically oriented, defaults to False
        
    t_tail : bool
        Flag indicating if wing is mounted on vertical tail, defaults to False
        
    taper : float
        Wing taper ratio, defaults to 0.0
        
    dihedral : float
        Wing dihedral angle, defaults to 0.0
        
    aspect_ratio : float
        Wing aspect ratio, defaults to 0.0
        
    thickness_to_chord : float
        Average thickness-to-chord ratio, defaults to 0.0
        
    aerodynamic_center : list
        Location of aerodynamic center [x, y, z], defaults to [0.0, 0.0, 0.0]
        
    exposed_root_chord_offset : float
        Offset of exposed root from centerline, defaults to 0.0
        
    total_length : float
        Total length of wing, defaults to 0.0
        
    spans : Data
        Collection of span measurements
        
        - projected : float
            Projected span, defaults to 0.0
        - total : float
            Total span including dihedral effects, defaults to 0.0
            
    areas : Data
        Collection of area measurements
        
        - reference : float
            Reference area, defaults to 0.0
        - exposed : float
            Exposed area, defaults to 0.0
        - affected : float
            Area affected by high-lift devices, defaults to 0.0
        - wetted : float
            Wetted area, defaults to 0.0
            
    chords : Data
        Collection of chord measurements
        
        - mean_aerodynamic : float
            Mean aerodynamic chord, defaults to 0.0
        - mean_geometric : float
            Mean geometric chord, defaults to 0.0
        - root : float
            Root chord, defaults to 0.0
        - tip : float
            Tip chord, defaults to 0.0
            
    sweeps : Data
        Collection of sweep angles
        
        - quarter_chord : float
            Quarter-chord sweep angle, defaults to 0.0
        - leading_edge : float
            Leading edge sweep angle, defaults to None
        - half_chord : float
            Half-chord sweep angle, defaults to 0.0
            
    twists : Data
        Collection of twist angles
        
        - root : float
            Root section twist angle, defaults to 0.0
        - tip : float
            Tip section twist angle, defaults to 0.0
            
    high_lift : bool
        Flag indicating presence of high-lift devices, defaults to False
        
    symbolic : bool
        Flag for symbolic computation mode, defaults to False
        
    high_mach : bool
        Flag for high Mach number flow, defaults to False
        
    vortex_lift : bool
        Flag for vortex lift modeling, defaults to False
        
    transition_x_upper : float
        Upper surface transition location, defaults to 0.0
        
    transition_x_lower : float
        Lower surface transition location, defaults to 0.0
        
    dynamic_pressure_ratio : float
        Local to freestream dynamic pressure ratio, defaults to 0.0
        
    Airfoil : Container
        Collection of airfoil definitions, initialized empty
        
    segments : Container
        Collection of wing segments, initialized empty
        
    control_surfaces : Container
        Collection of control surfaces, initialized empty
        
    fuel_tanks : Container
        Collection of fuel tanks, initialized empty

    Notes
    -----
    The Wing class serves as the foundation for all lifting surfaces in RCAIDE. 
    It provides:
    
    * Geometric definition capabilities
    * Segment management
    * Control surface integration
    * Fuel tank integration
    * Mass properties computation

    See Also
    --------
    RCAIDE.Library.Components.Wings.Main_Wing
        Primary lifting surface implementation
    RCAIDE.Library.Components.Wings.Horizontal_Tail
        Horizontal stabilizer implementation
    RCAIDE.Library.Components.Wings.Vertical_Tail
        Vertical stabilizer implementation
    """      

    def __defaults__(self):
        """
        Sets default values for the wing attributes.
        """         
        self.tag                               = 'wing'
        self.mass_properties                   = Mass_Properties()
        self.origin                            = np.array([[0.0,0.0,0.0]])
                                               
        self.symmetric                         = True
        self.vertical                          = False
        self.t_tail                            = False
        self.taper                             = 0.0
        self.dihedral                          = 0.0
        self.aspect_ratio                      = 0.0
        self.thickness_to_chord                = 0.0
        self.aerodynamic_center                = [0.0,0.0,0.0]
        self.exposed_root_chord_offset         = 0.0
        self.total_length                      = 0.0
                                               
        self.spans                             = Data()
        self.spans.projected                   = 0.0
        self.spans.total                       = 0.0
                                               
        self.areas                             = Data()
        self.areas.reference                   = 0.0
        self.areas.exposed                     = 0.0
        self.areas.affected                    = 0.0
        self.areas.wetted                      = 0.0
                                               
        self.chords                            = Data()
        self.chords.mean_aerodynamic           = 0.0
        self.chords.mean_geometric             = 0.0
        self.chords.root                       = 0.0
        self.chords.tip                        = 0.0
                                               
        self.sweeps                            = Data()
        self.sweeps.quarter_chord              = 0.0
        self.sweeps.leading_edge               = None
        self.sweeps.half_chord                 = 0.0        
                                               
        self.twists                            = Data()
        self.twists.root                       = 0.0
        self.twists.tip                        = 0.0
                                               
        self.high_lift                         = False
        self.symbolic                          = False 
        self.high_mach                         = False
        self.vortex_lift                       = False
                                               
        self.transition_x_upper                = 0.0
        self.transition_x_lower                = 0.0
                                               
        self.dynamic_pressure_ratio            = 0.0
                                               
        self.airfoil                           = None 
        
        self.segments                          = Container()
        self.control_surfaces                  = Container()
        self.fuel_tanks                        = Container()

    def append_segment(self, segment):
        """
        Adds a new segment to the wing's segment container.

        Parameters
        ----------
        segment : Data
            Wing segment to be added
        """
        # Assert database type
        if not isinstance(segment,RCAIDE.Library.Components.Wings.Segments.Segment):
            raise Exception('input component must be of type Segment')

        # Store data
        self.segments.append(segment)

        return
    
    def append_airfoil(self, airfoil):
        """
        Adds an airfoil definition to the wing.

        Parameters
        ----------
        airfoil : Data
            Airfoil data to be added
        """
        # Assert database type
        if not isinstance(airfoil,RCAIDE.Library.Components.Airfoils.Airfoil):
            raise Exception('input component must be of type Airfoil')

        # Store data
        self.airfoil = airfoil

        return        

    def append_control_surface(self, control_surface):
        """
        Adds a control surface to the wing.

        Parameters
        ----------
        control_surface : Data
            Control surface to be added
        """
        # Assert database type
        if not isinstance(control_surface,Data):
            raise Exception('input control surface must be of type Data()')

        # Store data
        self.control_surfaces.append(control_surface)

        return
    
    def append_fuel_tank(self, fuel_tank):
        """
        Adds a fuel tank to the wing.

        Parameters
        ----------
        fuel_tank : Data
            Fuel tank to be added
        """
        # Assert database type
        if not isinstance(fuel_tank,Data):
            raise Exception('input component must be of type Data()')

        # Store data
        self.fuel_tanks.append(fuel_tank)

        return
 
    def compute_moment_of_inertia(self, mass, center_of_gravity=[[0, 0, 0]], fuel_flag=False): 
        """
        Computes the moment of inertia tensor for the wing.

        Parameters
        ----------
        mass : float
            Wing mass
        center_of_gravity : list, optional
            Reference point coordinates, defaults to [[0, 0, 0]]
        fuel_flag : bool, optional
            Flag to include fuel mass, defaults to False

        Returns
        -------
        ndarray
            3x3 moment of inertia tensor
        """
        I = compute_wing_moment_of_inertia(self, mass, center_of_gravity, fuel_flag) 
        return I   
    
class Container(Component.Container):
    def get_children(self):
        """ Returns the components that can go inside
        
        Assumptions:
        None
    
        Source:
        N/A
    
        Inputs:
        None
    
        Outputs:
        None
    
        Properties Used:
        N/A
        """       
        from . import Main_Wing
        from . import Vertical_Tail
        from . import Horizontal_Tail
        
        return [Main_Wing,Vertical_Tail,Horizontal_Tail]


# ------------------------------------------------------------
#  Handle Linking
# ------------------------------------------------------------

Wing.Container = Container



# RCAIDE/Library/Methods/Stability/Moment_of_Inertia/__init__.py 

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .compute_wing_moment_of_inertia     import compute_wing_moment_of_inertia
from .compute_cuboid_moment_of_inertia   import compute_cuboid_moment_of_inertia
from .compute_fuselage_moment_of_inertia import compute_fuselage_moment_of_inertia
from .compute_cylinder_moment_of_inertia import compute_cylinder_moment_of_inertia
from .compute_aircraft_moment_of_inertia import compute_aircraft_moment_of_inertia


# RCAIDE/Methods/Energy/Propulsors/Networks/Turboprop_Propulsor/compute_turboprop_performance.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports      

from RCAIDE.Framework.Core                                           import Data 
from RCAIDE.Library.Methods.Propulsors.Converters.Ram                import compute_ram_performance
from RCAIDE.Library.Methods.Propulsors.Converters.Combustor          import compute_combustor_performance
from RCAIDE.Library.Methods.Propulsors.Converters.Compressor         import compute_compressor_performance
from RCAIDE.Library.Methods.Propulsors.Converters.Turbine            import compute_turbine_performance
from RCAIDE.Library.Methods.Propulsors.Converters.Expansion_Nozzle   import compute_expansion_nozzle_performance 
from RCAIDE.Library.Methods.Propulsors.Converters.Compression_Nozzle import compute_compression_nozzle_performance
from RCAIDE.Library.Methods.Propulsors.Turboprop_Propulsor           import compute_thrust
 
# python imports 
from   copy import deepcopy
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
# compute_turboprop_performance
# ---------------------------------------------------------------------------------------------------------------------- 
def compute_turboprop_performance(turboprop,state,center_of_gravity= [[0.0, 0.0,0.0]]):    
    
    ''' Computes the performance of one turboprop
    
    Assumptions: 
    N/A

    Source:
    N/A

    Inputs:  
    conditions           - operating conditions data structure     [-]  
    fuel_line            - fuel line                               [-] 
    turboprop            - turboprop data structure                [-] 
    total_power          - power of turboprop group                [W] 

    Outputs:  
    thrust               - thrust of turboprop group               [W]
    power                - power of turboprop group                [W] 
    stored_results_flag  - boolean for stored results              [-]     
    stored_propulsor_tag - name of turboprop with stored results   [-]
    
    Properties Used: 
    N.A.        
    ''' 
    conditions                                            = state.conditions 
    noise_conditions                                      = conditions.noise[turboprop.tag]  
    turboprop_conditions                                  = conditions.energy[turboprop.tag]
    
    ram                                                   = turboprop.ram
    inlet_nozzle                                          = turboprop.inlet_nozzle
    compressor                                            = turboprop.compressor
    combustor                                             = turboprop.combustor
    high_pressure_turbine                                 = turboprop.high_pressure_turbine
    low_pressure_turbine                                  = turboprop.low_pressure_turbine
    core_nozzle                                           = turboprop.core_nozzle  

    # unpack component conditions
    turboprop_conditions                                  = conditions.energy[turboprop.tag]
    ram_conditions                                        = turboprop_conditions[ram.tag]     
    inlet_nozzle_conditions                               = turboprop_conditions[inlet_nozzle.tag]
    core_nozzle_conditions                                = turboprop_conditions[core_nozzle.tag] 
    compressor_conditions                                 = turboprop_conditions[compressor.tag]  
    combustor_conditions                                  = turboprop_conditions[combustor.tag]
    lpt_conditions                                        = turboprop_conditions[low_pressure_turbine.tag]
    hpt_conditions                                        = turboprop_conditions[high_pressure_turbine.tag] 
     
    # Step 1: Set the working fluid to determine the fluid properties
    ram.working_fluid                                     = turboprop.working_fluid

    # Step 2: Compute flow through the ram , this computes the necessary flow quantities and stores it into conditions
    compute_ram_performance(ram,ram_conditions,conditions)

    # Step 3: link inlet nozzle to ram 
    inlet_nozzle_conditions.inputs.stagnation_temperature = ram_conditions.outputs.stagnation_temperature
    inlet_nozzle_conditions.inputs.stagnation_pressure    = ram_conditions.outputs.stagnation_pressure
    inlet_nozzle_conditions.inputs.static_temperature     = ram_conditions.outputs.static_temperature
    inlet_nozzle_conditions.inputs.static_pressure        = ram_conditions.outputs.static_pressure
    inlet_nozzle_conditions.inputs.mach_number            = ram_conditions.outputs.mach_number
    inlet_nozzle.working_fluid                            = ram.working_fluid

    # Step 4: Compute flow through the inlet nozzle
    compute_compression_nozzle_performance(inlet_nozzle,inlet_nozzle_conditions,conditions)      

    # Step 5: Link low pressure compressor to the inlet nozzle 
    compressor_conditions.inputs.stagnation_temperature   = inlet_nozzle_conditions.outputs.stagnation_temperature
    compressor_conditions.inputs.stagnation_pressure      = inlet_nozzle_conditions.outputs.stagnation_pressure
    compressor_conditions.inputs.static_temperature       = inlet_nozzle_conditions.outputs.static_temperature
    compressor_conditions.inputs.static_pressure          = inlet_nozzle_conditions.outputs.static_pressure
    compressor_conditions.inputs.mach_number              = inlet_nozzle_conditions.outputs.mach_number  
    compressor.working_fluid                              = inlet_nozzle.working_fluid 

    # Step 6: Compute flow through the low pressure compressor
    compute_compressor_performance(compressor,compressor_conditions,conditions)
    
    # Step 7: Link the combustor to the high pressure compressor
    combustor_conditions.inputs.stagnation_temperature    = compressor_conditions.outputs.stagnation_temperature
    combustor_conditions.inputs.stagnation_pressure       = compressor_conditions.outputs.stagnation_pressure
    combustor_conditions.inputs.static_temperature        = compressor_conditions.outputs.static_temperature
    combustor_conditions.inputs.static_pressure           = compressor_conditions.outputs.static_pressure
    combustor_conditions.inputs.mach_number               = compressor_conditions.outputs.mach_number  
    combustor.working_fluid                               = compressor.working_fluid 
    
    # Step 8: Compute flow through the high pressor compressor 
    compute_combustor_performance(combustor,combustor_conditions,conditions)

    #link the high pressure turbione to the combustor 
    hpt_conditions.inputs.stagnation_temperature          = combustor_conditions.outputs.stagnation_temperature
    hpt_conditions.inputs.stagnation_pressure             = combustor_conditions.outputs.stagnation_pressure
    hpt_conditions.inputs.fuel_to_air_ratio               = combustor_conditions.outputs.fuel_to_air_ratio 
    hpt_conditions.inputs.static_temperature              = combustor_conditions.outputs.static_temperature
    hpt_conditions.inputs.static_pressure                 = combustor_conditions.outputs.static_pressure
    hpt_conditions.inputs.mach_number                     = combustor_conditions.outputs.mach_number 
    hpt_conditions.inputs.compressor                      = compressor_conditions.outputs 
    high_pressure_turbine.working_fluid                   = combustor.working_fluid
    hpt_conditions.inputs.bypass_ratio                    = 0.0
    
    compute_turbine_performance(high_pressure_turbine,hpt_conditions,conditions)
            
    #link the low pressure turbine to the high pressure turbine 
    lpt_conditions.inputs.stagnation_temperature          = hpt_conditions.outputs.stagnation_temperature
    lpt_conditions.inputs.stagnation_pressure             = hpt_conditions.outputs.stagnation_pressure 
    lpt_conditions.inputs.static_temperature              = hpt_conditions.outputs.static_temperature
    lpt_conditions.inputs.static_pressure                 = hpt_conditions.outputs.static_pressure 
    lpt_conditions.inputs.mach_number                     = hpt_conditions.outputs.mach_number     
    lpt_conditions.inputs.compressor                      = Data()
    lpt_conditions.inputs.compressor.work_done            = 0.0        
    lpt_conditions.inputs.bypass_ratio                    = 0.0 
    lpt_conditions.inputs.fuel_to_air_ratio               = combustor_conditions.outputs.fuel_to_air_ratio 
    low_pressure_turbine.working_fluid                    = high_pressure_turbine.working_fluid    
     
    compute_turbine_performance(low_pressure_turbine,lpt_conditions,conditions)
    
    #link the core nozzle to the low pressure turbine
    core_nozzle_conditions.inputs.stagnation_temperature  = lpt_conditions.outputs.stagnation_temperature
    core_nozzle_conditions.inputs.stagnation_pressure     = lpt_conditions.outputs.stagnation_pressure
    core_nozzle_conditions.inputs.static_temperature      = lpt_conditions.outputs.static_temperature
    core_nozzle_conditions.inputs.static_pressure         = lpt_conditions.outputs.static_pressure  
    core_nozzle_conditions.inputs.mach_number             = lpt_conditions.outputs.mach_number   
    core_nozzle.working_fluid                             = low_pressure_turbine.working_fluid 
    
    #flow through the core nozzle
    compute_expansion_nozzle_performance(core_nozzle,core_nozzle_conditions,conditions)

    # compute the thrust using the thrust component
    
    turboprop_conditions.cpt                              = lpt_conditions.outputs.cp
    turboprop_conditions.R_t                              = lpt_conditions.outputs.gas_constant
    turboprop_conditions.stag_temp_lpt_out                = lpt_conditions.inputs.stagnation_temperature
    turboprop_conditions.stag_temp_lpt_in                 = lpt_conditions.outputs.stagnation_temperature
    turboprop_conditions.stag_temp_hpt_out                = hpt_conditions.inputs.stagnation_temperature
    turboprop_conditions.stag_temp_hpt_in                 = hpt_conditions.outputs.stagnation_temperature
    turboprop_conditions.stag_press_lpt_exit              = lpt_conditions.outputs.stagnation_pressure 
    turboprop_conditions.core_exit_velocity               = core_nozzle_conditions.outputs.velocity
    turboprop_conditions.core_area_ratio                  = core_nozzle_conditions.outputs.area_ratio
    turboprop_conditions.core_nozzle                      = core_nozzle_conditions.outputs
    turboprop_conditions.P9                               = core_nozzle_conditions.outputs.static_pressure    
    turboprop_conditions.T9                               = core_nozzle_conditions.outputs.static_temperature
    turboprop_conditions.cpc                              = compressor_conditions.outputs.cp
    turboprop_conditions.gamma_c                          = compressor_conditions.outputs.gamma
    turboprop_conditions.R_c                              = compressor_conditions.outputs.gas_constant 
    turboprop_conditions.total_temperature_reference      = compressor_conditions.inputs.stagnation_temperature
    turboprop_conditions.total_pressure_reference         = compressor_conditions.inputs.stagnation_pressure 
    turboprop_conditions.fuel_to_air_ratio                = combustor_conditions.outputs.fuel_to_air_ratio
    turboprop_conditions.combustor_stagnation_temperature = combustor_conditions.outputs.stagnation_temperature       
    
    # Compute the power
    compute_thrust(turboprop,turboprop_conditions,conditions) 

    # Compute forces and moments
    moment_vector      = 0*state.ones_row(3)
    thrust_vector      = 0*state.ones_row(3)
    thrust_vector[:,0] = turboprop_conditions.thrust[:,0]
    moment_vector[:,0] = turboprop.origin[0][0] -   center_of_gravity[0][0] 
    moment_vector[:,1] = turboprop.origin[0][1]  -  center_of_gravity[0][1] 
    moment_vector[:,2] = turboprop.origin[0][2]  -  center_of_gravity[0][2]
    M                  =  np.cross(moment_vector, thrust_vector)   
    moment             = M 
    power              = turboprop_conditions.power 
 

    # Store data
    core_nozzle_res = Data(
                exit_static_temperature             = core_nozzle_conditions.outputs.static_temperature,
                exit_static_pressure                = core_nozzle_conditions.outputs.static_pressure,
                exit_stagnation_temperature         = core_nozzle_conditions.outputs.stagnation_temperature,
                exit_stagnation_pressure            = core_nozzle_conditions.outputs.static_pressure,
                exit_velocity                       = core_nozzle_conditions.outputs.velocity
            )
  
    noise_conditions.turboprop.core_nozzle   = core_nozzle_res  
    
    # Pack results    
    stored_results_flag    = True
    stored_propulsor_tag   = turboprop.tag 
    
    return thrust_vector,moment,power,stored_results_flag,stored_propulsor_tag

def reuse_stored_turboprop_data(turboprop,state,network,stored_propulsor_tag,center_of_gravity= [[0.0, 0.0,0.0]]):
    '''Reuses results from one turboprop for identical propulsors
    
    Assumptions: 
    N/A

    Source:
    N/A

    Inputs:  
    conditions           - operating conditions data structure     [-]  
    fuel_line            - fuelline                                [-] 
    turboprop            - turboprop data structure              [-] 
    total_power          - power of turboprop group               [W] 

    Outputs:  
    total_power          - power of turboprop group               [W] 
    
    Properties Used: 
    N.A.        
    ''' 
    conditions                        = state.conditions  
    conditions.energy[turboprop.tag]  = deepcopy(conditions.energy[stored_propulsor_tag])
    conditions.noise[turboprop.tag]   = deepcopy(conditions.noise[stored_propulsor_tag])
    
    # compute moment  
    moment_vector      = 0*state.ones_row(3)
    thrust_vector      = 0*state.ones_row(3)
    thrust_vector[:,0] = conditions.energy[turboprop.tag].thrust[:,0] 
    moment_vector[:,0] = turboprop.origin[0][0] -   center_of_gravity[0][0] 
    moment_vector[:,1] = turboprop.origin[0][1]  -  center_of_gravity[0][1] 
    moment_vector[:,2] = turboprop.origin[0][2]  -  center_of_gravity[0][2]
    moment             = np.cross(moment_vector, thrust_vector)         
  
    power                                   = conditions.energy[turboprop.tag].power 
    conditions.energy[turboprop.tag].moment =  moment 
    return thrust_vector,moment,power    


# RCAIDE/Framework/External_Interfaces/OpenVSP/vsp_propeller.py

# Created:  Sep 2021, R. Erhard
# Modified:

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
# RCAIDE imports 
import RCAIDE
from RCAIDE.Framework.Core import Units , Data
import numpy as np
import scipy as sp
import string
try:
    import vsp as vsp
except ImportError:
    try:
        import openvsp as vsp
    except ImportError:
        # This allows RCAIDE to build without OpenVSP
        pass

# This enforces lowercase names
chars = string.punctuation + string.whitespace
t_table = str.maketrans( chars          + string.ascii_uppercase ,
                         '_'*len(chars) + string.ascii_lowercase )

# ---------------------------------------------------------------------------------------------------------------------- 
#  vsp read rotor
# ---------------------------------------------------------------------------------------------------------------------- 
def read_vsp_rotor(prop_id, units_type='SI',write_airfoil_file=True):
    """This reads an OpenVSP rotor geometry and writes it into a RCAIDE rotor format.

    Assumptions:
    1. Written for OpenVSP 3.24.0

    Source:
    N/A

    Inputs:
    1. VSP 10-digit geom ID for rotor.
    2. units_type set to 'SI' (default) or 'Imperial'.
    3. write_airfoil_file set to True (default) or False
    4. number_of_radial_stations is the radial discretization used to extract the rotor design from OpenVSP

    Outputs:
    Writes RCAIDE rotor object, with these geometries, from VSP:
    	rotor.
    		origin                                  [m] in all three dimensions
    		orientation				[deg] in all three dimensions
    		number_of_blades			[-]
    		tip_radius				[m]
    		hub_radius				[m]
    		twist_distribution			[deg]
    		chord_distribution			[m]
    		radius_distribution			[m]
    		sweep_distribution			[deg]
    		mid_chord_alignment			[m]
    		max_thickness_distribution		[m]
    		thickness_to_chord			[-]
    		blade_solidity				[-]
    		rotation			        [-]
    		thickness_to_chord                      [-]
                beta34                                  [radians]
                pre_cone                                [radians]
                rake                                    [radians]
                skew                                    [radians]
                axial                                   [radians]
                tangential                              [radians]
    		dihedral                                [radians]
    		symmetric                               <boolean>
    		tag                                     <string>
    		Segments.
    		  tag                                   <string>
    		  twist                                 [radians]
    		  percent_span_location                 [-]  .1 is 10%
    		  root_chord_percent                    [-]  .1 is 10%
    		  dihedral_outboard                     [radians]
    		  sweeps.quarter_chord                  [radians]
    		  thickness_to_chord                    [-]
    		  airfoil                               <NACA 4-series, 6 series, or airfoil file>

    Properties Used:
    N/A
    """


    # Check if this is a rotor or a lift rotor 
    rotor 	= RCAIDE.Library.Components.Propulsors.Converters.Rotor()

    # Set the units
    if units_type == 'SI':
        units_factor = Units.meter * 1.
    elif units_type == 'imperial':
        units_factor = Units.foot * 1.
    elif units_type == 'inches':
        units_factor = Units.inch * 1.

    # Apply a tag to the rotor
    if vsp.GetGeomName(prop_id):
        tag = vsp.GetGeomName(prop_id)
        tag = tag.translate(t_table)
        rotor.tag = tag
    else:
        rotor.tag = 'propgeom'


    scaling           = vsp.GetParmVal(prop_id, 'Scale', 'XForm')
    units_factor      = units_factor*scaling

    # Propeller location (absolute)
    rotor.origin 	= [[0.0,0.0,0.0]]
    rotor.origin[0][0] 	= vsp.GetParmVal(prop_id, 'X_Location', 'XForm') * units_factor
    rotor.origin[0][1] 	= vsp.GetParmVal(prop_id, 'Y_Location', 'XForm') * units_factor
    rotor.origin[0][2] 	= vsp.GetParmVal(prop_id, 'Z_Location', 'XForm') * units_factor

    # Propeller orientation
    rotor.orientation_euler_angles 	= [0.0,0.0,0.0]
    rotor.orientation_euler_angles[0] 	= vsp.GetParmVal(prop_id, 'X_Rotation', 'XForm') * Units.degrees
    rotor.orientation_euler_angles[1] 	= vsp.GetParmVal(prop_id, 'Y_Rotation', 'XForm') * Units.degrees
    rotor.orientation_euler_angles[2] 	= vsp.GetParmVal(prop_id, 'Z_Rotation', 'XForm') * Units.degrees

    # Get the rotor parameter IDs
    parm_id    = vsp.GetGeomParmIDs(prop_id)
    parm_names = []
    for i in range(len(parm_id)):
        parm_name = vsp.GetParmName(parm_id[i])
        parm_names.append(parm_name)

    # Run the vsp Blade Element analysis
    vsp.SetStringAnalysisInput( "BladeElement" , "PropID" , (prop_id,) )
    rid = vsp.ExecAnalysis( "BladeElement" )
    Nc  = len(vsp.GetDoubleResults(rid,"YSection_000"))

    rotor.vtk_airfoil_points           = 2*Nc
    rotor.CLi                          = vsp.GetParmVal(parm_id[parm_names.index('CLi')])
    rotor.blade_solidity               = vsp.GetParmVal(parm_id[parm_names.index('Solidity')])
    rotor.number_of_blades             = int(vsp.GetParmVal(parm_id[parm_names.index('NumBlade')]))

    rotor.tip_radius                   = vsp.GetDoubleResults(rid, "Diameter" )[0] / 2 * units_factor
    rotor.radius_distribution          = np.array(vsp.GetDoubleResults(rid, "Radius" )) * rotor.tip_radius

    rotor.radius_distribution[-1]      = 0.99 * rotor.tip_radius # BEMT requires max nondimensional radius to be less than 1.0
    if rotor.radius_distribution[0] == 0.:
        start = 1
        rotor.radius_distribution = rotor.radius_distribution[start:]
    else:
        start = 0

    rotor.hub_radius                   = rotor.radius_distribution[0]

    rotor.chord_distribution           = np.array(vsp.GetDoubleResults(rid, "Chord" ))[start:]  * rotor.tip_radius # vsp gives c/R
    rotor.twist_distribution           = np.array(vsp.GetDoubleResults(rid, "Twist" ))[start:]  * Units.degrees
    rotor.sweep_distribution           = np.array(vsp.GetDoubleResults(rid, "Sweep" ))[start:]
    rotor.mid_chord_alignment          = np.tan(rotor.sweep_distribution*Units.degrees)  * rotor.radius_distribution
    rotor.thickness_to_chord           = np.array(vsp.GetDoubleResults(rid, "Thick" ))[start:]
    rotor.max_thickness_distribution   = rotor.thickness_to_chord*rotor.chord_distribution * units_factor
    rotor.Cl_distribution              = np.array(vsp.GetDoubleResults(rid, "CLi" ))[start:]

    # Extra data from VSP BEM for future use in BEVW
    rotor.beta34                       = vsp.GetDoubleResults(rid, "Beta34" )[0]  # pitch at 3/4 radius
    rotor.pre_cone                     = vsp.GetDoubleResults(rid, "Pre_Cone")[0]
    rotor.rake                         = np.array(vsp.GetDoubleResults(rid, "Rake"))[start:]
    rotor.skew                         = np.array(vsp.GetDoubleResults(rid, "Skew"))[start:]
    rotor.axial                        = np.array(vsp.GetDoubleResults(rid, "Axial"))[start:]
    rotor.tangential                   = np.array(vsp.GetDoubleResults(rid, "Tangential"))[start:]

    # Set rotor rotation
    rotor.rotation = 1

    # ---------------------------------------------
    # Rotor Airfoil
    # ---------------------------------------------
    if write_airfoil_file:
        print("Airfoil write not yet implemented. Defaulting to NACA 4412 airfoil for rotor cross section.")

    return rotor

# ---------------------------------------------------------------------------------------------------------------------- 
# write_vsp_rotor_bem
# ---------------------------------------------------------------------------------------------------------------------- 
def write_vsp_rotor_bem(vsp_bem_filename,rotor):
    """   This functions writes a .bem file for OpenVSP
    Assumptions:
        None

    Source:
        None
    Inputs:
        OpenVSP .bem filename
        RCAIDE Propeller Data Structure
    Outputs:
        OpenVSP .bem file
    Properties Used:
        N/A
    """
    vsp_bem = open(vsp_bem_filename,'w')
    with open(vsp_bem_filename,'w') as vsp_bem:
        make_header_text(vsp_bem, rotor)

        make_section_text(vsp_bem,rotor)

        make_airfoil_text(vsp_bem,rotor)

    # Now import this rotor
    vsp.ImportFile(vsp_bem_filename,vsp.IMPORT_BEM,'')

    return


# ---------------------------------------------------------------------------------------------------------------------- 
# make_header_text
# ---------------------------------------------------------------------------------------------------------------------- 
def make_header_text(vsp_bem,rotor):
    """This function writes the header of the OpenVSP .bem file
    Assumptions:
        None

    Source:
        None
    Inputs:
        vsp_bem - OpenVSP .bem file
        rotor    - RCAIDE rotor data structure

    Outputs:
        NA
    Properties Used:
        N/A
    """
    header_base = \
'''...{0}...
Num_Sections: {1}
Num_Blade: {2}
Diameter: {3}
Beta 3/4 (deg): {4}
Feather (deg): 0.00000000
Pre_Cone (deg): 0.00000000
Center: {5}, {6}, {7}
Normal: {8}, {9}, {10}
'''
    # Unpack inputs
    name      = rotor.tag
    N         = len(rotor.radius_distribution)
    B         = int(rotor.number_of_blades)
    D         = np.round(rotor.tip_radius*2,5)
    beta      = np.round(rotor.twist_distribution/Units.degrees,5)
    X         = np.round(rotor.origin[0][0],5)
    Y         = np.round(rotor.origin[0][1],5)
    Z         = np.round(rotor.origin[0][2],5) 
    Xn        = np.round(rotor.orientation_euler_angles[2],5) / Units.degrees
    Yn        = np.round(rotor.orientation_euler_angles[0],5) / Units.degrees 
    Zn        = np.round(rotor.orientation_euler_angles[1],5) / Units.degrees

    beta_3_4  = np.interp(rotor.tip_radius*0.75,rotor.radius_distribution,beta)

    # Insert inputs into the template
    header_text = header_base.format(name,N,B,D,beta_3_4,X,Y,Z,Xn,Yn,Zn)
    vsp_bem.write(header_text)

    return

# ---------------------------------------------------------------------------------------------------------------------- 
# make_section_text
# ---------------------------------------------------------------------------------------------------------------------- 
def make_section_text(vsp_bem,rotor):
    """This function writes the sectional information of the rotor
    Assumptions:
        None

    Source:
        None
    Inputs:
        vsp_bem - OpenVSP .bem file
        rotor    - RCAIDE rotor data structure

    Outputs:
        NA
    Properties Used:
        N/A
    """
    header = \
        '''Radius/R, Chord/R, Twist (deg), Rake/R, Skew/R, Sweep, t/c, CLi, Axial, Tangential\n'''

    N          = len(rotor.radius_distribution)
    r_R        = np.zeros(N)
    c_R        = np.zeros(N)
    r_R        = rotor.radius_distribution/rotor.tip_radius
    c_R        = rotor.chord_distribution/rotor.tip_radius
    beta_deg   = rotor.twist_distribution/Units.degrees
    Rake_R     = np.zeros(N)
    Skew_R     = np.zeros(N)
    Sweep      = np.arctan(rotor.mid_chord_alignment/rotor.radius_distribution)
    t_c        = rotor.thickness_to_chord
    
    if type(rotor) == RCAIDE.Library.Components.Propulsors.Converters.Lift_Rotor: 
        CLi        = np.ones(N)*rotor.hover.design_Cl  
    elif type(rotor) == RCAIDE.Library.Components.Propulsors.Converters.Propeller:
        CLi        = np.ones(N)*rotor.cruise.design_Cl   
    elif type(rotor) == RCAIDE.Library.Components.Propulsors.Converters.Prop_Rotor: 
        CLi        = np.ones(N)*rotor.hover.design_Cl  
    
    Axial      = np.zeros(N)
    Tangential = np.zeros(N)

    # Write rotor station imformation
    vsp_bem.write(header)
    for i in range(N):
        section_text = format(r_R[i], '.7f')+ ", " + format(c_R[i], '.7f')+ ", " + format(beta_deg[i], '.7f')+ ", " +\
            format( Rake_R[i], '.7f')+ ", " + format(Skew_R[i], '.7f')+ ", " + format(Sweep[i], '.7f')+ ", " +\
            format(t_c[i], '.7f')+ ", " + format(CLi[i], '.7f') + ", "+ format(Axial[i], '.7f') + ", " +\
            format(Tangential[i], '.7f') + "\n"
        vsp_bem.write(section_text)

    return

# ---------------------------------------------------------------------------------------------------------------------- 
# make_airfoil_text
# ---------------------------------------------------------------------------------------------------------------------- 
def make_airfoil_text(vsp_bem,rotor):
    """This function writes the airfoil geometry into the vsp file
    Assumptions:
        None

    Source:
        None
    Inputs:
        vsp_bem - OpenVSP .bem file
        rotor    - RCAIDE rotor data structure

    Outputs:
        NA
    Properties Used:
        N/A
    """

    N             = len(rotor.radius_distribution)
    airfoils      = rotor.airfoils
    airfoil_list  = list(airfoils.keys())
    a_sec         = rotor.airfoil_polar_stations
    for i in range(N):
        airfoil_station_header = '\nSection ' + str(i) + ' X, Y\n'
        vsp_bem.write(airfoil_station_header)
        airfoil_index =  a_sec[i]
        airfoil       = airfoils[airfoil_list[airfoil_index]]
        airfoil_x     = airfoil.geometry.x_coordinates 
        airfoil_y     = airfoil.geometry.y_coordinates 

        for j in range(len(airfoil_x)):
            section_text = format(airfoil_x[j], '.7f')+ ", " + format(airfoil_y[j], '.7f') + "\n"
            vsp_bem.write(section_text)
    return



# RCAIDE/Library/Missions/Common/Unpack_Unknowns/energy.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  Unpack Unknowns
# ---------------------------------------------------------------------------------------------------------------------- 
def unknowns(segment):  
    ACV_T  =  segment.assigned_control_variables.throttle
    ACV_TA =  segment.assigned_control_variables.thrust_vector_angle
    
    for network in segment.analyses.energy.vehicle.networks: 
        if 'throttle' in segment: 
            for propulsor in network.propulsors: 
                segment.state.conditions.energy[propulsor.tag].throttle[:,0] = segment.throttle
            
        if ACV_T.active: 
            for i in range(len(ACV_T.assigned_propulsors)): 
                propulsor_group = ACV_T.assigned_propulsors[i]
                for propulsor_name in propulsor_group:  
                    segment.state.conditions.energy[propulsor_name].throttle = segment.state.unknowns["throttle_" + str(i)]  
    
       # Thrust Vector Control 
        if ACV_TA.active:                
            for i in range(len(ACV_TA.assigned_propulsors)): 
                propulsor_group = ACV_TA.assigned_propulsors[i]
                for propulsor_name in propulsor_group:  
                    segment.state.conditions.energy[propulsor_name].commanded_thrust_vector_angle = segment.state.unknowns["thrust_vector_" + str(i)]
    return 
     
 
    



# RCAIDE/Library/Methods/Weights/Correlation_Buildups/Human_Powered/compute_tail_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  Tail Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_tail_weight(Sts,bts,cts,Ntsr,t_cts,qm):        
    """ Compute weight of human-powered aircraft tail     
    
    Assumptions:
       All of this is from AIAA 89-2048, units are in kg. These weight estimates
       are from the MIT Daedalus and are valid for very lightweight
       carbon fiber composite structures. This may need to be solved iteratively since
       gross weight is an input.
       
    Source: 
        MIT Daedalus
                
    Inputs:
        Sts -      tail surface area                                [meters]
        bts -      tail surface span                                [meters]
        cts -      average tail surface chord                       [meters]
        deltats -  average rib spacing to average chord ratio       [dimensionless]
        Ntsr -     number of tail surface ribs (bts^2)/(deltats*Sts)[dimensionless]
        t_cts -    tail airfoil thickness to chord ratio            [dimensionless]
        qm -       dynamic pressure at maneuvering speed            [Pascals]
    
    Outputs:
        Wtss -     weight of tail surface spar                      [kilogram]
        Wtsr -     weight of tail surface ribs                      [kilogram]
        WtsLE -    weight of tail surface leading edge              [kilogram]
        Wtsc -     weight of tail surface covering                  [kilogram]
            
    Properties Used:
        N/A
    """     
    deltats = (bts**2)/(Sts*Ntsr)
    
    #Rudder & Elevator Primary Structure:
    Wtss = (bts * 4.15e-2 + (bts**2) * 3.91e-3) * (1.0 + ((qm * Sts)/78.5 - 1.0)/12.0)
    
    #Rudder & Elevator Secondary Structure:
    Wtsr = Ntsr * (cts**2 * t_cts * 1.16e-1 + cts * 4.01e-3)
    Wts  = 0.174*((Sts**2)*(deltats**(4./3.))/bts)
    Wtsc = Sts * 1.93e-2
    
    weight = Wtss + Wtsr + Wts + Wtsc
    
    return weight


# RCAIDE/Library/Methods/Geometry/Two_Dimensional/Airfoil/compute_airfoil_properties.py
# 
# 
# Created:  Jul 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------    

# RCAIDE imports 
import RCAIDE
from RCAIDE.Framework.Core                                                          import Data , Units 
from RCAIDE.Library.Methods.Aerodynamics.Airfoil_Panel_Method.airfoil_analysis      import airfoil_analysis
from RCAIDE.Library.Methods.Geometry.Airfoil.import_airfoil_polars                  import import_airfoil_polars 
from RCAIDE.Library.Methods.Geometry.Airfoil.compute_naca_4series                   import compute_naca_4series  
from RCAIDE.Library.Methods.Aerodynamics.AERODAS.pre_stall_coefficients             import pre_stall_coefficients
from RCAIDE.Library.Methods.Aerodynamics.AERODAS.post_stall_coefficients            import post_stall_coefficients

# numpy imports 
import numpy as np
from scipy.interpolate     import RegularGridInterpolator
from RCAIDE.Framework.Core import interp2d 

# ----------------------------------------------------------------------------------------------------------------------
#  compute_airfoil_properties
# ----------------------------------------------------------------------------------------------------------------------   
def compute_airfoil_properties(airfoil_geometry, airfoil_polar_files = None,use_pre_stall_data=True):
    """This computes the aerodynamic properties and coefficients of an airfoil in stall regimes using pre-stall
    characterstics and AERODAS formation for post stall characteristics. This is useful for 
    obtaining a more accurate prediction of wing and blade loading as well as aeroacoustics. Pre stall characteristics 
    are obtained in the form of a text file of airfoil polar data obtained from airfoiltools.com
    
    Assumptions:
        None 
        
    Source
        None
        
    Inputs:
    airfoil_geometry                        <data_structure>
    airfoil_polar_files                     <string>
    boundary_layer_files                    <string>
    use_pre_stall_data                      [Boolean]
    Outputs:
    airfoil_data.
        cl_polars                           [unitless]
        cd_polars                           [unitless]      
        aoa_sweep                           [unitless]
         
    
    Properties Used:
    N/A
    """     
    Airfoil_Data   = Data()  
   
    # ----------------------------------------------------------------------------------------
    # Compute airfoil boundary layers properties 
    # ----------------------------------------------------------------------------------------   
    Airfoil_Data   = compute_boundary_layer_properties(airfoil_geometry,Airfoil_Data)
    num_polars     = len(Airfoil_Data.re_from_polar)
     
    # ----------------------------------------------------------------------------------------
    # Compute extended cl and cd polars 
    # ----------------------------------------------------------------------------------------    
    if airfoil_polar_files != None : 
        # check number of polars per airfoil in batch
        num_polars   = 0 
        n_p = len(airfoil_polar_files)
        if n_p < 3:
            raise AttributeError('Provide three or more airfoil polars to compute surrogate') 
        num_polars = max(num_polars, n_p)         
         
        # read in polars from files overwrite panel code  
        airfoil_file_data                          = import_airfoil_polars(airfoil_polar_files)  
        Airfoil_Data.aoa_from_polar                = airfoil_file_data.aoa_from_polar
        Airfoil_Data.re_from_polar                 = airfoil_file_data.re_from_polar 
        Airfoil_Data.lift_coefficients             = airfoil_file_data.lift_coefficients
        Airfoil_Data.drag_coefficients             = airfoil_file_data.drag_coefficients 
        
    # Get all of the coefficients for AERODAS wings
    AoA_sweep_deg         = np.linspace(-14,90,105)
    AoA_sweep_rad         = AoA_sweep_deg*Units.degrees    
    
    # Create an infinite aspect ratio wing
    geometry              = RCAIDE.Library.Components.Wings.Wing()
    geometry.aspect_ratio = np.inf
    geometry.section      = Data()  
    
    # AERODAS   
    CL = np.zeros((num_polars,len(AoA_sweep_deg)))
    CD = np.zeros((num_polars,len(AoA_sweep_deg)))  
    for j in range(num_polars):    
        airfoil_cl  = Airfoil_Data.lift_coefficients[j,:]
        airfoil_cd  = Airfoil_Data.drag_coefficients[j,:]   
        airfoil_aoa = Airfoil_Data.aoa_from_polar[j,:]/Units.degrees 
    
        # compute airfoil cl and cd for extended AoA range 
        CL[j,:],CD[j,:] = compute_extended_polars(airfoil_cl,airfoil_cd,airfoil_aoa,AoA_sweep_deg,geometry,use_pre_stall_data)  
         
    # ----------------------------------------------------------------------------------------
    # Store data 
    # ----------------------------------------------------------------------------------------    
    Airfoil_Data.reynolds_numbers    = Airfoil_Data.re_from_polar
    Airfoil_Data.angle_of_attacks    = AoA_sweep_rad 
    Airfoil_Data.lift_coefficients   = CL 
    Airfoil_Data.drag_coefficients   = CD    
        
    return Airfoil_Data
 
def compute_extended_polars(airfoil_cl,airfoil_cd,airfoil_aoa,AoA_sweep_deg,geometry,use_pre_stall_data): 
    """ Computes the aerodynamic polars of an airfoil over an extended angle of attack range
    
    Assumptions:
    Uses AERODAS formulation for post stall characteristics 
    Source:
    Models of Lift and Drag Coefficients of Stalled and Unstalled Airfoils in Wind Turbines and Wind Tunnels
    by D Spera, 2008
        
    Inputs:
    airfoil_cl          [unitless]
    airfoil_cd          [unitless]
    airfoil_aoa         [degrees]
    AoA_sweep_deg       [unitless]
    geometry            [N/A]
    use_pre_stall_data  [boolean]
    Outputs: 
    CL                  [unitless]
    CD                  [unitless]       
    
    Properties Used:
    N/A
    """    
    
    # Create dummy settings and state
    settings                                              = Data()
    state                                                 = Data()
    state.conditions                                      = Data()
    state.conditions.aerodynamics                         = Data()
    state.conditions.aerodynamics.pre_stall_coefficients  = Data()
    state.conditions.aerodynamics.post_stall_coefficients = Data()  
    
    # computing approximate zero lift aoa
    AoA_sweep_radians    = AoA_sweep_deg*Units.degrees
    airfoil_cl_plus      = airfoil_cl[airfoil_cl>0]
    idx_zero_lift        = np.where(airfoil_cl == min(airfoil_cl_plus))[0][0]
    airfoil_cl_crossing  = airfoil_cl[idx_zero_lift-1:idx_zero_lift+1]
    airfoil_aoa_crossing = airfoil_aoa[idx_zero_lift-1:idx_zero_lift+1]
    try:
        A0  = np.interp(0,airfoil_cl_crossing, airfoil_aoa_crossing)* Units.deg 
    except:
        A0 = airfoil_aoa[idx_zero_lift] * Units.deg 

    # max lift coefficent and associated aoa
    CL1max                  = np.max(airfoil_cl)
    idx_aoa_max_prestall_cl = np.where(airfoil_cl == CL1max)[0][0]
    ACL1                    = airfoil_aoa[idx_aoa_max_prestall_cl] * Units.degrees

    # computing approximate lift curve slope
    linear_idxs             = [int(np.where(airfoil_aoa==0)[0]),int(np.where(airfoil_aoa==4)[0])]
    cl_range                = airfoil_cl[linear_idxs]
    aoa_range               = airfoil_aoa[linear_idxs] * Units.degrees
    S1                      = (cl_range[1]-cl_range[0])/(aoa_range[1]-aoa_range[0])

    # max drag coefficent and associated aoa
    CD1max                  = np.max(airfoil_cd) 
    idx_aoa_max_prestall_cd = np.where(airfoil_cd == CD1max)[0][0]
    ACD1                    = airfoil_aoa[idx_aoa_max_prestall_cd] * Units.degrees     
    
    # Find the point of lowest drag and the CD
    idx_CD_min              = np.where(airfoil_cd==min(airfoil_cd))[0][0]
    ACDmin                  = airfoil_aoa[idx_CD_min] * Units.degrees
    CDmin                   = airfoil_cd[idx_CD_min]    
    
    # Setup data structures for this run
    ones                                                      = np.ones_like(AoA_sweep_radians)
    settings.section_zero_lift_angle_of_attack                = A0
    state.conditions.aerodynamics.angle_of_attack             = AoA_sweep_radians* ones  
    geometry.section.angle_attack_max_prestall_lift           = ACL1 * ones 
    geometry.pre_stall_maximum_drag_coefficient_angle         = ACD1 * ones 
    geometry.pre_stall_maximum_lift_coefficient               = CL1max * ones 
    geometry.pre_stall_maximum_lift_drag_coefficient          = CD1max * ones 
    geometry.section.minimum_drag_coefficient                 = CDmin * ones 
    geometry.section.minimum_drag_coefficient_angle_of_attack = ACDmin
    geometry.pre_stall_lift_curve_slope                       = S1
    
    # Get prestall coefficients
    CL1, CD1 = pre_stall_coefficients(state,settings,geometry)
    
    # Get poststall coefficents
    CL2, CD2 = post_stall_coefficients(state,settings,geometry)
    
    # Take the maxes
    CL_ij = np.fmax(CL1,CL2)
    CL_ij[AoA_sweep_radians<=A0] = np.fmin(CL1[AoA_sweep_radians<=A0],CL2[AoA_sweep_radians<=A0])
    
    CD_ij = np.fmax(CD1,CD2)
    
    # Pack this loop
    CL = CL_ij
    CD = CD_ij
    
    if use_pre_stall_data == True:
        CL, CD = apply_pre_stall_data(AoA_sweep_deg, airfoil_aoa, airfoil_cl, airfoil_cd, CL, CD)
        
    return CL,CD

def apply_pre_stall_data(AoA_sweep_deg, airfoil_aoa, airfoil_cl, airfoil_cd, CL, CD): 
    '''Applies prestall data to lift and drag curve slopes
    
    Assumptions:
    None
    
    Source:
    None
    
    Inputs:
    AoA_sweep_deg  [degrees]
    airfoil_aoa    [degrees]
    airfoil_cl     [unitless]
    airfoil_cd     [unitless]
    CL             [unitless] 
    CD             [unitless]
    
    Outputs
    CL             [unitless]
    CD             [unitless] 
    
    
    Properties Used:
    N/A
    
    '''

    # Coefficients in pre-stall regime taken from experimental data:
    aoa_locs    = (AoA_sweep_deg>=airfoil_aoa[0]) * (AoA_sweep_deg<=airfoil_aoa[-1])
    aoa_in_data = AoA_sweep_deg[aoa_locs]
    
    # if the data is within experimental use it, if not keep the surrogate values
    CL[aoa_locs] = airfoil_cl[abs(aoa_in_data[:,None] - airfoil_aoa[None,:]).argmin(axis=-1)]
    CD[aoa_locs] = airfoil_cd[abs(aoa_in_data[:,None] - airfoil_aoa[None,:]).argmin(axis=-1)]
    
    # remove kinks/overlap between pre- and post-stall                
    data_lb       = np.where(CD == airfoil_cd[0])[0][0]
    data_ub       = np.where(CD == airfoil_cd[-1])[0][-1]
    CD[0:data_lb] = np.maximum(CD[0:data_lb], CD[data_lb]*np.ones_like(CD[0:data_lb]))
    CD[data_ub:]  = np.maximum(CD[data_ub:],  CD[data_ub]*np.ones_like(CD[data_ub:])) 
    
    return CL, CD
 
def compute_boundary_layer_properties(airfoil_geometry,Airfoil_Data): 
    '''Computes the boundary layer properties of an airfoil for a sweep of Reynolds numbers 
    and angle of attacks. 
    
    Source:
    None
    
    Assumptions:
    None 
    
    Inputs:
    airfoil_geometry   <data_structure>
    Airfoil_Data       <data_structure>
    
    Outputs:
    Airfoil_Data       <data_structure>
    
    Properties Used:
    N/A
    '''
    if airfoil_geometry == None:
        print('No airfoil defined, NACA 0012 surrogates will be used') 
        a_names                       = ['0012']                
        airfoil_geometry              = compute_naca_4series(a_names, npoints= 100)    
    
    AoA_sweep = np.array([-4,0,2,4,8,10,14])*Units.degrees 
    Re_sweep  = np.array([1,5,10,30,50,75,100])*1E4  
    AoA_vals  = np.tile(AoA_sweep[None,:],(len(Re_sweep) ,1))
    Re_vals   = np.tile(Re_sweep[:,None],(1, len(AoA_sweep)))     
    
    # run airfoil analysis  
    af_res  = airfoil_analysis(airfoil_geometry,AoA_vals,Re_vals) 
    
    # store data
    Airfoil_Data.aoa_from_polar                                     = np.tile(AoA_sweep[None,:],(len(Re_sweep),1)) 
    Airfoil_Data.re_from_polar                                      = Re_sweep 
    Airfoil_Data.lift_coefficients                                  = af_res.cl_invisc
    Airfoil_Data.drag_coefficients                                  = af_res.cd_invisc
    Airfoil_Data.cm                                                 = af_res.cm_invisc
    Airfoil_Data.boundary_layer                                     = Data() 
    Airfoil_Data.boundary_layer.angle_of_attacks                    = AoA_sweep 
    Airfoil_Data.boundary_layer.reynolds_numbers                    = Re_sweep
    fL                                                              = np.transpose(af_res.fL, axes=[1,2,0])
    fD                                                              = np.transpose(af_res.fD, axes=[1,2,0]) 
    Airfoil_Data.lift_distribution_func                             = RegularGridInterpolator((AoA_sweep,Re_sweep),fL,method = 'linear',   bounds_error=False, fill_value=None)
    Airfoil_Data.drag_distribution_func                             = RegularGridInterpolator((AoA_sweep,Re_sweep),fD,method = 'linear',   bounds_error=False, fill_value=None)
    Airfoil_Data.lift_distribution                                  = fL 
    Airfoil_Data.drag_distribution                                  = fD  
    
    return Airfoil_Data


# RCAIDE/Library/Components/Propulsors/Propuslor.py
#  
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

# RCAIDE imports  
from RCAIDE.Library.Components                   import Component 

# ----------------------------------------------------------------------------------------------------------------------
#  Propusor
# ---------------------------------------------------------------------------------------------------------------------- 
class Propulsor(Component):
    """
    Base class for all propulsion system components in RCAIDE.
    
    Attributes
    ----------
    tag : str
        Identifier for the propulsor, defaults to 'propulsor'
        
    active : bool
        Flag indicating if the propulsor is operational, defaults to True
        
    wing_mounted : bool
        Flag indicating if the propulsor is mounted on a wing, defaults to True
    
    Notes
    -----
    This class serves as the foundation for all propulsion system implementations 
    in RCAIDE. It provides the basic structure and common attributes needed for:
    - Electric propulsion systems (rotors, ducted fans)
    - Internal combustion engine systems (fixed and constant-speed propellers)
    - Gas turbine systems (turbofans, turbojets, turboprops)
    
    The class inherits from Component, providing basic component functionality
    while adding propulsion-specific features. Derived classes must implement
    their own performance calculation methods and system-specific attributes.
    
    **Definitions**

    'Propulsor'
        Any device that generates thrust or lift force through the manipulation 
        of fluid momentum (air in most cases)
    
    'Wing-mounted'
        Configuration where the propulsor is attached to the aircraft's wing 
        rather than the fuselage or other structures
    
    See Also
    --------
    RCAIDE.Library.Components.Component
    RCAIDE.Library.Components.Propulsors.Electric_Rotor
    RCAIDE.Library.Components.Propulsors.Electric_Ducted_Fan
    RCAIDE.Library.Components.Propulsors.ICE_Propeller
    RCAIDE.Library.Components.Propulsors.Constant_Speed_ICE_Propeller
    """
    
    
    def __defaults__(self):
        """ This sets the default values.
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            None
        """          
        self.tag                          = 'propulsor' 
        self.active                       = True 
        self.wing_mounted                 = True 
        
    


# RCAIDE/Compoments/Landing_Gear/Main_Landing_Gear.py
# 
# Created:  Nov 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports     
from .Landing_Gear import Landing_Gear

# ---------------------------------------------------------------------------------------------------------------------- 
#  Main_Landing_Gear
# ---------------------------------------------------------------------------------------------------------------------- 
class Main_Landing_Gear(Landing_Gear):
    """
    Main landing gear component that handles the primary loads during landing, takeoff, 
    and ground operations.

    Attributes
    ----------
    tag : str
        Unique identifier for the main landing gear component, defaults to 'main_gear'
        
    units : float
        Number of main landing gear units in the assembly, defaults to 0
        
    strut_length : float
        Length of the main gear strut assembly, defaults to 0
        
    tire_diameter : float
        Diameter of the main gear tires, defaults to 0
        
    wheels : float
        Number of wheels per main landing gear unit, defaults to 0

    Notes
    -----
    The main landing gear typically carries 80-90% of the aircraft weight and is 
    designed to handle the primary impact loads during landing.
    
    **Major Assumptions**
    
    * Symmetrical arrangement about aircraft centerline
    * All units in an assembly have identical properties
    * Load distribution is uniform across all wheels in a unit
    
    **Definitions**

    'Multi-Bogey'
        Configuration with multiple wheels per strut to distribute loads
        
    'Track Width'
        Lateral distance between main gear units, critical for aircraft stability

    See Also
    --------
    RCAIDE.Library.Components.Landing_Gear.Landing_Gear
        Base landing gear class
    RCAIDE.Library.Components.Landing_Gear.Nose_Landing_Gear
        Nose gear component that works in conjunction with main gear
    """

    def __defaults__(self):
        """
        Sets default values for the main landing gear attributes.
        """
        self.tag           = 'main_gear'
        self.units         = 0. # number of main landing gear units        
        self.strut_length  = 0.
        self.tire_diameter = 0. 
        self.wheels        = 0. # number of wheels on the main landing gear 


# RCAIDE/Library/Missions/Common/Pre_Process/energy.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  energy
# ----------------------------------------------------------------------------------------------------------------------  
def energy(mission):
    """ Appends all unknows and residuals to the network 
    
        Assumptions:
            N/A
        
        Inputs:
            None
            
        Outputs:
            None 

        Properties Used:
        N/A                
    """       
    for segment in mission.segments:
        for network in segment.analyses.energy.vehicle.networks:
            network.add_unknowns_and_residuals_to_segment(segment) 
    return 


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/General_Aviation/compute_horizontal_tail_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE 
from RCAIDE.Framework.Core import  Units
import  numpy as  np

# ----------------------------------------------------------------------------------------------------------------------
# Horizontal Tail Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_horizontal_tail_weight(S_h, AR_h, sweep_h, q_c, taper_h, t_c_h,Nult,TOW):       
    """ 
        Calculate the weight of the horizontal tail for a General Aviation-type aircraft
        
        Source:
            Aircraft Design: A Conceptual Approach by Raymer (pg 460 in 4th edition)
        
      
        Inputs:
            S_h  = trapezoidal area of horizontal tail           [meters**2]
            Ar_h = aspect ratio of horizontal tail               [dimensionless]
            sweep_h = quarter chord sweep of the horizontal tail [radians]
            q_c  = dynamic pressure at cruise                    [Pascals]
            
            
        Outputs:
            weight - weight of the horizontal tail               [kilograms]
            
        Assumptions:
            calculated weight of the horizontal tail including the elevator
            Assume that the elevator is 25% of the horizontal tail
        
    """
    # unpack inputs
    W_0   = TOW / Units.lb # Convert kg to lbs
    S_ht  = S_h/ Units.ft**2 # Convert from meters squared to ft squared  
    q     = q_c /(Units.force_pound / Units.ft**2.)

    #Calculate weight of wing for traditional aircraft horizontal tail
    weight_English = .016*((Nult*W_0)**.414)*(q**.168)*(S_ht**.896)*((100.*t_c_h/np.cos(sweep_h))**(-.12))*((AR_h/(np.cos(sweep_h)**2))**.043)*(taper_h**(-.02))
    weight         = weight_English * Units.lbs # Convert from lbs to kg

    return weight


## @ingroup Analyses-Propulsion
# Momentum_Theory_Wake.py
#
# Created:  Jan 2022, R. Erhard
# Modified: 

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------
from RCAIDE.Framework.Core import Data
from RCAIDE.Library.Components import Component 
from RCAIDE.Library.Components import Wings
from RCAIDE.Library.Methods.Propulsors.Converters.Rotor.Wake.Momentum_Theory_Wake.fidelity_zero_wake_convergence import fidelity_zero_wake_convergence
from RCAIDE.Library.Methods.Propulsors.Converters.Rotor.Wake.Momentum_Theory_Wake.compute_fidelity_zero_induced_velocity import compute_fidelity_zero_induced_velocity 
from RCAIDE.Library.Methods.Aerodynamics.Vortex_Lattice_Method.extract_wing_collocation_points import extract_wing_collocation_points

import numpy as np
# ----------------------------------------------------------------------
#  Generalized Rotor Class
# ----------------------------------------------------------------------
## @ingroup Analyses-Propulsion
class Momentum_Theory_Wake(Component):
    """This is a general rotor wake component. 

    Assumptions:
    None

    Source:
    None
    """
    def __defaults__(self):
        """This sets the default values for the component to function.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        None

        Outputs:
        None

        Properties Used:
        None
        """

        self.tag            = 'rotor_wake'
        self.wake_method    = 'Fidelity_Zero'

    
    def evaluate(self,rotor,wake_inputs,conditions):
        """
        
        Wake evaluation is performed using a simplified vortex wake method for Fidelity Zero, 
        following Helmholtz vortex theory.
        
        Assumptions:
        None

        Source:
        Drela, M. "Qprop Formulation", MIT AeroAstro, June 2006
        http://web.mit.edu/drela/Public/web/qprop/qprop_theory.pdf

        Inputs:
           self         - rotor wake
           rotor        - RCAIDE rotor
           wake_inputs.
              Ua        - Axial velocity
              Ut        - Tangential velocity
              r         - radius distribution
           conditions   - conditions
           
           
        Outputs:
           va  - axially-induced velocity from rotor wake
           vt  - tangentially-induced velocity from rotor wake
        
        Properties Used:
        None
        
        
        """
        
        va, vt = fidelity_zero_wake_convergence(self, rotor, wake_inputs)
            
        return va, vt
    
    def evaluate_slipstream(self,rotor,rotor_conditions,geometry,ctrl_pts,wing_instance=None):
        """
        Evaluates the velocities induced by the rotor on a specified wing of the vehicle.
        If no wing instance is specified, uses main wing or last available wing in geometry.
        
        Assumptions:
        None

        Source:
        N/A

        Inputs:
           self         - rotor wake
           rotor        - rotor
           geometry     - vehicle geometry
           
        Outputs:
           wake_V_ind   - induced velocity from rotor wake at (VD.XC, VD.YC, VD.ZC)
        
        Properties Used:
        None
        """
        # Check for wing if wing instance is unspecified
        if wing_instance == None:
            nmw = 0
            # check for main wing
            for i,wing in enumerate(geometry.wings):
                if not isinstance(wing,Wings.Main_Wing): continue
                nmw +=1                
                wing_instance = wing
                wing_instance_idx = i
            if nmw == 1:
                pass
            elif nmw>1:
                print("No wing specified for slipstream analysis. Multiple main wings in vehicle, using the last one.")
            else:
                print("No wing specified for slipstream analysis. No main wing defined, using the last wing in vehicle.")
                wing_instance = wing 
                wing_instance_idx = i
        
        # Isolate the VD components corresponding to this wing instance
        wing_CPs, slipstream_vd_ids = extract_wing_collocation_points(geometry, wing_instance_idx)
        
        # Evaluate rotor slipstream effect on specified wing instance
        rot_V_wake_ind = self.evaluate_wake_velocities(rotor,rotor_conditions,wing_CPs,ctrl_pts)
        
        # Expand
        wake_V_ind = np.zeros((ctrl_pts,geometry.vortex_distribution.n_cp,3))
        wake_V_ind[:,slipstream_vd_ids,:] = rot_V_wake_ind
        
            
        return wake_V_ind
    
    def evaluate_wake_velocities(self,rotor,rotor_conditions,evaluation_points,ctrl_pts):
        """
        Links the rotor wake to compute the wake-induced velocities at the specified
        evaluation points.
        
        Assumptions:
        None

        Source:
        N/A

        Inputs:
           self               - rotor wake
           rotor              - rotor
           evaluation_points  - points at which to evaluate the rotor wake-induced velocities 
           
        Outputs:
           prop_V_wake_ind  - induced velocity from rotor wake at (VD.XC, VD.YC, VD.ZC)
        
        Properties Used:
        None
        """  
         
        rot_V_wake_ind = compute_fidelity_zero_induced_velocity(rotor,rotor_conditions,evaluation_points,ctrl_pts)  
        
        return rot_V_wake_ind
    


# RCAIDE/Framework/Analyses/Emissions/Emission_Index_Correlation_Method.py
#  
# Created:  Jul 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
from RCAIDE.Framework.Analyses    import Process 
from RCAIDE.Library.Methods.Emissions.Emission_Index_Empirical_Method import *  
from .Emissions                   import Emissions 
  

# ----------------------------------------------------------------------------------------------------------------------
#  Correlation_Buildup
# ---------------------------------------------------------------------------------------------------------------------- 
class Emission_Index_Correlation_Method(Emissions): 
    """
    Analysis class for computing aircraft emissions using empirical correlation methods.

    Attributes
    ----------
    process : Process
        Container for computational processes
        - compute : Process
            Computation process container
            - emissions : function
                Function handle for emissions calculation

    Notes
    -----
    This class implements an emissions analysis method based on empirical correlations
    and pre-defined emission indices.

    The analysis follows these steps:
    1. Initialization of computation processes
    2. Assignment of empirical correlation functions
    3. Evaluation of emissions for given mission segments

    **Major Assumptions**
    * Emission indices are constant for each fuel type
    * Linear scaling with fuel flow rate
    * Well-mixed combustion products
    * Steady-state operation

    **Definitions**

    'Emission Index (EI)'
        Mass of pollutant emitted per unit mass of fuel burned [g/kg_fuel]

    'Process'
        Container for computational methods and their settings

    See Also
    --------
    RCAIDE.Framework.Analyses.Emissions : Parent emissions analysis class
    RCAIDE.Library.Methods.Emissions.Emission_Index_Empirical_Method : Implementation methods

    References
    ----------
    [1] Norman, P. D., et al. (2003). Development of the technical basis for a new emissions 
        parameter covering the whole AIRcraft operation: NEPAIR (ENV.C.1/ETU/2000/0094). 
        Final Technical Report.
    [2] Dopelheuer, A., & Lecht, M. (1999). Influence of engine performance on emission 
        characteristics. RTO MP-14, paper 20.
    """    
    
    def __defaults__(self): 
        """
        This sets the default values and methods for the analysis.
    
        Assumptions:
        None
    
        Source:
        None 
        """ 

        # build the evaluation process
        compute                         = Process()  
        compute.emissions               = None  
        self.process                    = Process()
        self.process.compute            = compute 
                
        return
            
    def initialize(self):   
        """
        This function computes the emissions of different species from the Emission Indices 
        available in literature.
    
        Assumptions:
        None
    
        Source:
        None 
        """         
        
        compute   =  self.process.compute     
        compute.emissions  = evaluate_correlation_emissions_indices
        return 


    def evaluate(self,segment):
        """
        The default evaluate function.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        None

        Outputs:
        results   <RCAIDE data class>

        Properties Used:
        self.settings
        self.vehicle
        """          
        settings = self.settings
        vehicle  = self.vehicle   
        results  = self.process.compute(segment,settings,vehicle)

        return results             


# cryocooler_model.py
# 
# Created:  Feb 2020,   K.Hamilton
# Modified: Nov 2021,   S. Claridge

# ----------------------------------------------------------------------
#  Imports
# ---------------------------------------------------------------------- 
# package imports 
import numpy as np

# ----------------------------------------------------------------------
#  Cryocooler Model 
# ----------------------------------------------------------------------     
def cryocooler_model(Cryocooler):

    """ Calculate the power required by the cryocooler based on the cryocooler type, the required cooling power, and the temperature conditions.

    Assumptions:
        Based on mass data for Cryomech cryocoolers as per the datasheets for ground based non-massreduced coolers available via the cryomech website: https://www.cryomech.com/cryocoolers/.
        The mass is calculated for the requested power level, the cryocooler should be sized for the maximum power level required as its mass will not change during the flight.
        The efficiency scales with required cooling power and temperature only.
        The temperature difference and efficiency are taken not to scale with ambient temperature. This should not matter in the narrow range of temperatures in which aircraft operate, i.e. for ambient temperatures between -50 and 50 C.
        
    Source: 
        https://www.cryomech.com/cryocoolers/
        
    Inputs:
        self.inputs
            cryo_temp -         cryogenic output temperature required                                   [kelvin]
            cooling_power -     cooling power required of the cryocooler                                [watts]
            

    Outputs:
        self.inputs
            rated_power -   electrical input power required by the cryocooler         [watts]
            mass -          mass of the cryocooler and supporting components          [kilogram]

    Properties Used:
        self.
            cooler_type -       cryocooler type used
            amb_temp -          ambient temperature the cooler will reject heat to, defaults to 19C     [kelvin]
"""      

    amb_temp        = Cryocooler.ambient_temp
    cooling_power   = Cryocooler.inputs.cooling_power
    cryo_temp       = Cryocooler.inputs.cryo_temp 


    # Prevent unrealistic temperature changes.
    if np.amin(cryo_temp) < 1.:

        cryo_temp = np.maximum(cryo_temp, 5.)
        print("Warning: Less than zero kelvin not possible, setting cryogenic temperature target to 5K.")

    # Warn if ambient temperature is very low.
    if np.amin(amb_temp) < 200.:

        print("Warning: Suprisingly low ambient temperature, check altitude.")

    # Calculate the shift in achievable minimum temperature based on the the ambient temperature (temp_amb) and the datasheet operating temperature (19C, 292.15K)
    temp_offset = 292.15 - amb_temp

    # Calculate the required temperature difference the cryocooler must produce.
    temp_diff = amb_temp-cryo_temp

    # Disable if the target temperature is greater than the ambient temp. Technically cooling like this is possible, however there are better cooling technologies to use if this is the required scenario.
    if np.amin(temp_diff) < 0.:

        temp_diff = np.maximum(temp_diff, 0.)
        print("Warning: Temperature conditions are not well suited to cryocooler use. Cryocooler disabled.")

    # Set the parameters of the cooler based on the cooler type and the operating conditions. The default ambient operating temperature (19C) is used as a base.
    if Cryocooler.cooler_type ==   'fps': #Free Piston Stirling

        temp_minRT      =  35.0      # Minimum temperature achievable by this type of cooler when rejecting to an ambient temperature of 19C (K)
        temp_min        =  temp_minRT - temp_offset   # Updated minimum achievable temperature based on the supplied ambient temperature (K)
        eff             =  0.0014*(cryo_temp-temp_min) # Efficiency function. This is a line fit from a survey of Cryomech coolers in November 2019  
        input_power     =  cooling_power/eff           # Electrical input power (W)
        mass            =  0.0098*input_power+1.0769   # Total cooler mass function. Fit from November 2019 Cryomech data. (kg)

    elif Cryocooler.cooler_type == 'GM': #Gifford McMahon

        temp_minRT      =  5.4
        temp_min        =  temp_minRT - temp_offset
        eff             =  0.0005*(cryo_temp-temp_min)
        input_power     =  cooling_power/eff
        mass            =  0.0129*input_power+63.08 

    elif Cryocooler.cooler_type == 'sPT': #Single Pulsetube

        temp_minRT      =  16.0
        temp_min        =  temp_minRT - temp_offset
        eff             =  0.0002*(cryo_temp-temp_min)
        input_power     =  cooling_power/eff
        mass            =  0.0079*input_power+51.124  

    elif Cryocooler.cooler_type == 'dPT': #Double Pulsetube

        temp_minRT      =  8.0
        temp_min        =  temp_minRT - temp_offset
        eff             =  0.00001*(cryo_temp-temp_min)
        input_power     =  cooling_power/eff
        mass            =  0.0111*input_power+73.809  

    else:

        print("Warning: Unknown Cryocooler type")
        return[0.0,0.0]

    # Warn if the cryogenic temperature is unachievable
    diff = cryo_temp - temp_min

    if np.amin(diff) < 0.0:

        eff         =   0.0
        input_power =   None
        mass        =   None

        print("Warning: The required cryogenic temperature of " + str(cryo_temp) + " is not achievable using a " + Cryocooler.cooler_type + " cryocooler at an ambient temperature of " + str(amb_temp) + ". The minimum temperature achievable is " + str(temp_min))


    return [input_power, mass]





# RCAIDE/Library/Methods/Propulsors/Converters/expansion_nozzle/append_expansion_nozzle_conditions.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke  

from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
# append_expansion_nozzle_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_expansion_nozzle_conditions(expansion_nozzle,segment,propulsor_conditions):   
    propulsor_conditions[expansion_nozzle.tag]                      = Conditions()
    propulsor_conditions[expansion_nozzle.tag].inputs               = Conditions()
    propulsor_conditions[expansion_nozzle.tag].outputs              = Conditions() 
    return 


# RCAIDE/Library/Methods/Energy/Propulsors/Turboshaft_Propulsor/design_turboprop.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  

# RCAIDE Imports     
import RCAIDE
from RCAIDE.Framework.Core                                           import Data
from RCAIDE.Framework.Mission.Common                                 import Conditions
from RCAIDE.Library.Methods.Propulsors.Converters.Ram                import compute_ram_performance
from RCAIDE.Library.Methods.Propulsors.Converters.Combustor          import compute_combustor_performance
from RCAIDE.Library.Methods.Propulsors.Converters.Compressor         import compute_compressor_performance
from RCAIDE.Library.Methods.Propulsors.Converters.Turbine            import compute_turbine_performance
from RCAIDE.Library.Methods.Propulsors.Converters.Expansion_Nozzle   import compute_expansion_nozzle_performance 
from RCAIDE.Library.Methods.Propulsors.Converters.Compression_Nozzle import compute_compression_nozzle_performance
from RCAIDE.Library.Methods.Propulsors.Turboprop_Propulsor           import size_core 
from RCAIDE.Library.Methods.Propulsors.Common                        import compute_static_sea_level_performance 

# Python package imports   
import numpy                                                                as np

# ----------------------------------------------------------------------------------------------------------------------  
#  Design Turboshaft
# ----------------------------------------------------------------------------------------------------------------------   
def design_turboprop(turboprop):  
    #check if mach number and temperature are passed
    if(turboprop.design_mach_number==None or turboprop.design_altitude==None):
        
        #raise an error
        raise NameError('The sizing conditions require an altitude and a Mach number')
    
    else:
        #call the atmospheric model to get the conditions at the specified altitude
        atmosphere                                        = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
        atmo_data                                         = atmosphere.compute_values(turboprop.design_altitude,turboprop.design_isa_deviation)
        planet                                            = RCAIDE.Library.Attributes.Planets.Earth()
                                                          
        p                                                 = atmo_data.pressure          
        T                                                 = atmo_data.temperature       
        rho                                               = atmo_data.density          
        a                                                 = atmo_data.speed_of_sound    
        mu                                                = atmo_data.dynamic_viscosity   
    
        # setup conditions
        conditions                                        = RCAIDE.Framework.Mission.Common.Results()
    
        # freestream conditions    
        conditions.freestream.altitude                    = np.atleast_1d(turboprop.design_altitude)
        conditions.freestream.mach_number                 = np.atleast_1d(turboprop.design_mach_number)
        conditions.freestream.pressure                    = np.atleast_1d(p)
        conditions.freestream.temperature                 = np.atleast_1d(T)
        conditions.freestream.density                     = np.atleast_1d(rho)
        conditions.freestream.dynamic_viscosity           = np.atleast_1d(mu)
        conditions.freestream.gravity                     = np.atleast_1d(planet.compute_gravity(turboprop.design_altitude))
        conditions.freestream.isentropic_expansion_factor = np.atleast_1d(turboprop.working_fluid.compute_gamma(T,p))
        conditions.freestream.Cp                          = np.atleast_1d(turboprop.working_fluid.compute_cp(T,p))
        conditions.freestream.R                           = np.atleast_1d(turboprop.working_fluid.gas_specific_constant)
        conditions.freestream.speed_of_sound              = np.atleast_1d(a)
        conditions.freestream.velocity                    = np.atleast_1d(a*turboprop.design_mach_number)
         
         
    fuel_line                                             = RCAIDE.Library.Components.Energy.Distributors.Fuel_Line()
    segment                                               = RCAIDE.Framework.Mission.Segments.Segment()  
    segment.state.conditions                              = conditions
    segment.state.conditions.energy[fuel_line.tag]        = Conditions()
    segment.state.conditions.noise[fuel_line.tag]         = Conditions()
    turboprop.append_operating_conditions(segment) 
    for tag, item in  turboprop.items(): 
        if issubclass(type(item), RCAIDE.Library.Components.Component):
            item.append_operating_conditions(segment,turboprop) 
         
    ram                                                   = turboprop.ram
    inlet_nozzle                                          = turboprop.inlet_nozzle
    compressor                                            = turboprop.compressor
    combustor                                             = turboprop.combustor
    high_pressure_turbine                                 = turboprop.high_pressure_turbine
    low_pressure_turbine                                  = turboprop.low_pressure_turbine
    core_nozzle                                           = turboprop.core_nozzle  

    # unpack component conditions
    turboprop_conditions                                  = conditions.energy[turboprop.tag]
    ram_conditions                                        = turboprop_conditions[ram.tag]     
    inlet_nozzle_conditions                               = turboprop_conditions[inlet_nozzle.tag]
    core_nozzle_conditions                                = turboprop_conditions[core_nozzle.tag] 
    compressor_conditions                                 = turboprop_conditions[compressor.tag]  
    combustor_conditions                                  = turboprop_conditions[combustor.tag]
    lpt_conditions                                        = turboprop_conditions[low_pressure_turbine.tag]
    hpt_conditions                                        = turboprop_conditions[high_pressure_turbine.tag] 
     
    # Step 1: Set the working fluid to determine the fluid properties
    ram.working_fluid                                     = turboprop.working_fluid

    # Step 2: Compute flow through the ram , this computes the necessary flow quantities and stores it into conditions
    compute_ram_performance(ram,ram_conditions,conditions)

    # Step 3: link inlet nozzle to ram 
    inlet_nozzle_conditions.inputs.stagnation_temperature = ram_conditions.outputs.stagnation_temperature
    inlet_nozzle_conditions.inputs.stagnation_pressure    = ram_conditions.outputs.stagnation_pressure
    inlet_nozzle_conditions.inputs.static_temperature     = ram_conditions.outputs.static_temperature
    inlet_nozzle_conditions.inputs.static_pressure        = ram_conditions.outputs.static_pressure
    inlet_nozzle_conditions.inputs.mach_number            = ram_conditions.outputs.mach_number
    inlet_nozzle.working_fluid                            = ram.working_fluid

    # Step 4: Compute flow through the inlet nozzle
    compute_compression_nozzle_performance(inlet_nozzle,inlet_nozzle_conditions,conditions)      

    # Step 5: Link low pressure compressor to the inlet nozzle 
    compressor_conditions.inputs.stagnation_temperature   = inlet_nozzle_conditions.outputs.stagnation_temperature
    compressor_conditions.inputs.stagnation_pressure      = inlet_nozzle_conditions.outputs.stagnation_pressure
    compressor_conditions.inputs.static_temperature       = inlet_nozzle_conditions.outputs.static_temperature
    compressor_conditions.inputs.static_pressure          = inlet_nozzle_conditions.outputs.static_pressure
    compressor_conditions.inputs.mach_number              = inlet_nozzle_conditions.outputs.mach_number  
    compressor.working_fluid                              = inlet_nozzle.working_fluid 

    # Step 6: Compute flow through the low pressure compressor
    compute_compressor_performance(compressor,compressor_conditions,conditions)
    
    # Step 7: Link the combustor to the high pressure compressor
    combustor_conditions.inputs.stagnation_temperature    = compressor_conditions.outputs.stagnation_temperature
    combustor_conditions.inputs.stagnation_pressure       = compressor_conditions.outputs.stagnation_pressure
    combustor_conditions.inputs.static_temperature        = compressor_conditions.outputs.static_temperature
    combustor_conditions.inputs.static_pressure           = compressor_conditions.outputs.static_pressure
    combustor_conditions.inputs.mach_number               = compressor_conditions.outputs.mach_number  
    combustor.working_fluid                               = compressor.working_fluid 
    
    # Step 8: Compute flow through the high pressor compressor 
    compute_combustor_performance(combustor,combustor_conditions,conditions)

    #link the high pressure turbione to the combustor 
    hpt_conditions.inputs.stagnation_temperature          = combustor_conditions.outputs.stagnation_temperature
    hpt_conditions.inputs.stagnation_pressure             = combustor_conditions.outputs.stagnation_pressure
    hpt_conditions.inputs.fuel_to_air_ratio               = combustor_conditions.outputs.fuel_to_air_ratio 
    hpt_conditions.inputs.static_temperature              = combustor_conditions.outputs.static_temperature
    hpt_conditions.inputs.static_pressure                 = combustor_conditions.outputs.static_pressure
    hpt_conditions.inputs.mach_number                     = combustor_conditions.outputs.mach_number 
    hpt_conditions.inputs.compressor                      = compressor_conditions.outputs 
    high_pressure_turbine.working_fluid                   = combustor.working_fluid
    hpt_conditions.inputs.bypass_ratio                    = 0.0
    
    compute_turbine_performance(high_pressure_turbine,hpt_conditions,conditions)
            
    #link the low pressure turbine to the high pressure turbine 
    lpt_conditions.inputs.stagnation_temperature          = hpt_conditions.outputs.stagnation_temperature
    lpt_conditions.inputs.stagnation_pressure             = hpt_conditions.outputs.stagnation_pressure 
    lpt_conditions.inputs.static_temperature              = hpt_conditions.outputs.static_temperature
    lpt_conditions.inputs.static_pressure                 = hpt_conditions.outputs.static_pressure 
    lpt_conditions.inputs.mach_number                     = hpt_conditions.outputs.mach_number     
    lpt_conditions.inputs.compressor                      = Data()
    lpt_conditions.inputs.compressor.work_done            = 0.0     
    lpt_conditions.inputs.bypass_ratio                    = 0.0    
    lpt_conditions.inputs.fuel_to_air_ratio               = combustor_conditions.outputs.fuel_to_air_ratio 
    low_pressure_turbine.working_fluid                    = high_pressure_turbine.working_fluid    
     
    compute_turbine_performance(low_pressure_turbine,lpt_conditions,conditions)
    
    #link the core nozzle to the low pressure turbine
    core_nozzle_conditions.inputs.stagnation_temperature  = lpt_conditions.outputs.stagnation_temperature
    core_nozzle_conditions.inputs.stagnation_pressure     = lpt_conditions.outputs.stagnation_pressure
    core_nozzle_conditions.inputs.static_temperature      = lpt_conditions.outputs.static_temperature
    core_nozzle_conditions.inputs.static_pressure         = lpt_conditions.outputs.static_pressure  
    core_nozzle_conditions.inputs.mach_number             = lpt_conditions.outputs.mach_number   
    core_nozzle.working_fluid                             = low_pressure_turbine.working_fluid 
    
    #flow through the core nozzle
    compute_expansion_nozzle_performance(core_nozzle,core_nozzle_conditions,conditions)

    # compute the thrust using the thrust component

    turboprop_conditions.cpt                              = lpt_conditions.outputs.cp
    turboprop_conditions.R_t                              = lpt_conditions.outputs.gas_constant
    turboprop_conditions.stag_temp_lpt_out                = lpt_conditions.inputs.stagnation_temperature
    turboprop_conditions.stag_temp_lpt_in                 = lpt_conditions.outputs.stagnation_temperature
    turboprop_conditions.stag_temp_hpt_out                = hpt_conditions.inputs.stagnation_temperature
    turboprop_conditions.stag_temp_hpt_in                 = hpt_conditions.outputs.stagnation_temperature
    turboprop_conditions.stag_press_lpt_exit              = lpt_conditions.outputs.stagnation_pressure 
    turboprop_conditions.core_exit_velocity               = core_nozzle_conditions.outputs.velocity
    turboprop_conditions.core_area_ratio                  = core_nozzle_conditions.outputs.area_ratio
    turboprop_conditions.core_nozzle                      = core_nozzle_conditions.outputs
    turboprop_conditions.P9                               = core_nozzle_conditions.outputs.static_pressure    
    turboprop_conditions.T9                               = core_nozzle_conditions.outputs.static_temperature
    turboprop_conditions.cpc                              = compressor_conditions.outputs.cp
    turboprop_conditions.gamma_c                          = compressor_conditions.outputs.gamma
    turboprop_conditions.R_c                              = compressor_conditions.outputs.gas_constant 
    turboprop_conditions.total_temperature_reference      = compressor_conditions.inputs.stagnation_temperature
    turboprop_conditions.total_pressure_reference         = compressor_conditions.inputs.stagnation_pressure 
    turboprop_conditions.fuel_to_air_ratio                = combustor_conditions.outputs.fuel_to_air_ratio
    turboprop_conditions.combustor_stagnation_temperature = combustor_conditions.outputs.stagnation_temperature 
    
    # Step 25: Size the core of the turboprop  
    size_core(turboprop,turboprop_conditions,conditions)
    
    # Step 26: Static Sea Level Thrust 
    compute_static_sea_level_performance(turboprop)

    turboprop.design_thrust_specific_fuel_consumption = turboprop_conditions.thrust_specific_fuel_consumption  
    turboprop.design_non_dimensional_thrust           = turboprop_conditions.non_dimensional_thrust            
    turboprop.design_core_mass_flow_rate              = turboprop_conditions.core_mass_flow_rate               
    turboprop.design_fuel_flow_rate                   = turboprop_conditions.fuel_flow_rate                     
    turboprop.design_power                            = turboprop_conditions.power                             
    turboprop.design_specific_power                   = turboprop_conditions.specific_power                    
    turboprop.design_power_specific_fuel_consumption  = turboprop_conditions.power_specific_fuel_consumption   
    turboprop.design_thermal_efficiency               = turboprop_conditions.thermal_efficiency                
    turboprop.design_propulsive_efficiency            = turboprop_conditions.propulsive_efficiency             
    return      
  